# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module defining constants referenced by methods and functions defined in  other modules in the package. This module
is not directly referenced by end users.
"""
from typing import Literal
__all__ = ['api_prefix', 'argus_prefix', 'DATETIME_COL_FMT', 'DEPLOY_TYPE', 'EXPECTED_DELIVERY', 'WORK_ORDER_PRIORITY',
           'WORK_ORDER_STATUS', 'WORK_ORDER_CATEGORY', 'ERROR_TYPE', 'MAPPING_ENTITIES', 'QUEUE_NAME', 'QUERY_LANGUAGE',
           'ACCOUNT', 'PROCESS_STATUS', 'AUTH_ENDPOINT_DEV', 'AUTH_ENDPOINT_PROD', 'SCHEDULE_TIMEZONE',
           'SWITCH_ENVIRONMENT', 'INTEGRATION_SETTINGS_EDITORS', 'DATA_INGESTION_CONTAINER', 'CACHE_SCOPE', 'TAG_LEVEL']

# @deprecated suggestion: Up for deprecation as this is not used anymore in the API
# api_prefix = "http://localhost:7071/api/SwitchApi/"
api_prefix = "https://pivotstreams.azurewebsites.net/api/SwitchApi/"

# @deprecated suggestion: Up for deprecation as this is not used anymore in the API (except for automation.reserve_instance)
# argus_prefix = "http://localhost:7071/api/"
argus_prefix = "https://arguslogicv4b.azurewebsites.net/api/"

AUTH_ENDPOINT_DEV = "https://restapi-dev.switchautomation.com/auth"
AUTH_ENDPOINT_PROD = "https://restapi-us.switchautomation.com/auth"

DATETIME_COL_FMT = Literal['DateTime', 'Date', 'Time']

# ['DateTime in 1 column',
#  'Date and Time in 2 columns',
#  'Split year and month in 2 columns',
#  'Start and End Date in 2 columns']

EXPECTED_DELIVERY = Literal['5min', '15min', 'Hourly', 'Daily', 'Weekly', 'Monthly', 'Quarterly']

# deploy_type 'File' for (FTP. Email, Upload)
DEPLOY_TYPE = Literal['Email', 'Ftp', 'Upload', 'Timer']

ACCOUNT = Literal['SwitchStorage', 'SwitchContainer', 'Argus', 'DataIngestion', 'GatewayMqttStorage']

DATA_INGESTION_CONTAINER = Literal['data-ingestion-adx', 'data-ingestion-timeseries-adx']

QUERY_LANGUAGE = Literal['sql', 'kql']

ERROR_TYPE = Literal['DateTime', 'MissingDevices', 'NonNumeric', 'MissingRequiredField(s)', 'MissingSite(s)',
                     'DuplicateRecords', 'MissingSensors', 'UnableToReadFile', 'InvalidFileType', 'InvalidInputSettings']

PROCESS_STATUS = Literal['ActionRequired', 'Failed']

MAPPING_ENTITIES = Literal['Installations', 'Devices/Sensors', 'Readings', 'Work Orders']

WORK_ORDER_CATEGORY = Literal['Preventative Maintenance', 'Tenant Request']

WORK_ORDER_PRIORITY = Literal['Low', 'Medium', 'High']

WORK_ORDER_STATUS = Literal['Submitted', 'Open', 'In Progress', 'Waiting for 3rd Party', 'Resolved', 'Abandoned',
                            'Closed']

QUEUE_NAME = Literal['task', 'highpriority']

SCHEDULE_TIMEZONE = Literal['Local', 'Utc']

SWITCH_ENVIRONMENT = Literal['Development', 'Staging', 'Testing', 'Production']

SUPPORT_PAYLOAD_TYPE = Literal['Sites', 'Sensors', 'Portfolio']

INTEGRATION_SETTINGS_EDITORS = Literal['text_box', 'numeric_stepper', 'custom_combo', 'tag_groups_combo',
                                       'equipment_combo']

CACHE_SCOPE = Literal['Portfolio', 'Task', 'DataFeed']
TAG_LEVEL = Literal['Site', 'Device', 'Sensor']

DATA_SET_QUERY_PARAMETER_TYPES = Literal['String', 'Number', 'DateTime', 'Keyword', 'Boolean', 'NumberArray', 'StringArray']



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for storing and retrieving state stored data in the Switch Platform.
"""

from .cache import (set_cache, get_cache)

__all__ = ['set_cache', 'get_cache']


# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for integrating asset creation, asset updates, data ingestion, etc into the Switch Automation Platform.
"""
import gc
import re
import sys
from typing import Union
import pandas
import pandas as pd
from ..integration.helpers import get_timezones
import pandera
import requests
import datetime
import logging
import uuid
from .._utils._platform import _get_structure, Blob
from .._utils._utils import (ApiInputs, _with_func_attrs, _column_name_cap, _work_order_schema, convert_bytes, IngestionMode)
from .._utils._constants import TAG_LEVEL
from ..integration._utils import _timezone_offsets, _upsert_entities_affected_count, _adx_support, _timezone_dst_offsets

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


@_with_func_attrs(df_required_columns=['ApiProjectId', 'InstallationId', 'NetworkDeviceId', 'UserId', 'BatchNo',
                                       'DriverUniqueId', 'Timestamp', 'DiscoveredValue', 'DriverClassName', 'JobId'],
                  df_optional_columns=['DriverDeviceType', 'ObjectPropertyTemplateName', 'UnitOfMeasureAbbrev',
                                       'DeviceName', 'DisplayName', 'EquipmentType', 'EquipmentLabel', ])
def upsert_discovered_records(df: pandas.DataFrame, api_inputs: ApiInputs, discovery_properties_columns: list,
                              device_tag_columns: list = None, sensor_tag_columns: list = None,
                              metadata_columns: list = None):  # , ontology_tag_columns: list = None):
    """Upsert discovered records to populate Build - Discovery & Selection UI.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe containing the discovered records including the minimum required set of columns.
    api_inputs : ApiInputs
        Object returned by initialize() function.
    discovery_properties_columns : list
        List of the discovery property columns returned by 3rd party API.
    device_tag_columns : list, default = None
        List of columns that represent device-level tag group(s) (Default value = None)
    sensor_tag_columns : list, default = None
        List of column names in input `df` that represent sensor-level tag group(s) (Default value = None).
    metadata_columns : list, default = None
        List of column names in input `df` that represent device-level metadata key(s) (Default value = None).
    # ontology_tag_columns : list, default = None
    #     List of BRICK schema or Haystack tags that apply to a given point (Default value = None).

    Returns
    -------
    tuple[pandas.DataFrame, pandas.DataFrame]
        (response_df, errors_df) - Returns the response dataframe and the dataframe containing the parsed errors text
        (if no errors, then empty dataframe).

    """

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    data_frame = df.copy()

    required_columns = ['ApiProjectId', 'InstallationId', 'NetworkDeviceId', 'UserId', 'BatchNo', 'DriverUniqueId',
                        'Timestamp', 'DiscoveredValue', 'DriverClassName', 'JobId']
    optional_columns = ['DriverDeviceType', 'ObjectPropertyTemplateName', 'UnitOfMeasureAbbrev', 'DeviceName',
                        'DisplayName', 'EquipmentType', 'EquipmentLabel', ]
    proposed_columns = data_frame.columns.tolist()

    if not set(required_columns).issubset(proposed_columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'integration.upsert_discovered_records() - df must contain the following columns: ' + ', '.join(
            required_columns) + '. Optional Columns include: ' + ', '.join(optional_columns)

    missing_optional_columns = set(optional_columns) - set(proposed_columns)
    for missing_column in missing_optional_columns:
        data_frame[missing_column] = None

    if discovery_properties_columns is not None and not set(discovery_properties_columns).issubset(data_frame.columns):
        logger.exception('Missing expected discovery property column(s): %s',
                         set(discovery_properties_columns).difference(proposed_columns))
        return 'Integration.upsert_discovered_records(): data_frame must contain the following discovery property column(s): ' + \
               ', '.join(discovery_properties_columns)
    elif discovery_properties_columns is None:
        logger.exception('Missing expected discovery property column(s): %s',
                         set(discovery_properties_columns).difference(proposed_columns))
        return 'Integration.upsert_discovered_records(): data_frame must contain the following discovery property' \
               ' columns: ' + ', '.join(discovery_properties_columns)

    if device_tag_columns is not None and not set(device_tag_columns).issubset(data_frame.columns):
        logger.exception('Missing expected device tag column(s): %s',
                         set(device_tag_columns).difference(proposed_columns))
        return 'Integration.upsert_discovered_records(): data_frame expected to contain the following device tag ' \
               'column(s): ' + ', '.join(device_tag_columns)
    elif device_tag_columns is None:
        device_tag_columns = []

    if sensor_tag_columns is not None and not set(sensor_tag_columns).issubset(data_frame.columns):
        logger.exception('Missing expected sensor tag column(s): %s',
                         set(sensor_tag_columns).difference(proposed_columns))
        return 'Integration.upsert_discovered_records(): data_frame expected to contain the following sensor tag ' \
               'column(s): ' + ', '.join(sensor_tag_columns)
    elif sensor_tag_columns is None:
        sensor_tag_columns = []

    if metadata_columns is not None and not set(metadata_columns).issubset(data_frame.columns):
        logger.exception('Missing expected metadata column(s): %s',
                         set(metadata_columns).difference(proposed_columns))
        return 'Integration.upsert_discovered_records(): data_frame expected to contain the following metadata ' \
               'column(s): ' + ', '.join(metadata_columns)
    elif metadata_columns is None:
        metadata_columns = []

    # if ontology_tag_columns is not None and not set(ontology_tag_columns).issubset(data_frame.columns):
    #     logger.exception('Missing expected ontology tag column(s): %s',
    #                      set(ontology_tag_columns).difference(proposed_columns))
    #     return 'Integration.upsert_discovered_records(): data_frame expected to contain the following ontology tag ' \
    #            'column(s): ' + ', '.join(ontology_tag_columns)
    # elif ontology_tag_columns is None:
    #     ontology_tag_columns = []

    # convert timestamp format to required format
    data_frame.Timestamp = data_frame.Timestamp.dt.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    data_frame = data_frame.rename(columns={'DiscoveredValue': 'CurrentValue'})

    column_dict = {'DiscoveryProperties': discovery_properties_columns + ['Timestamp'],
                   'DeviceTags': device_tag_columns,
                   'SensorTags': sensor_tag_columns,
                   'Metadata': metadata_columns,
                   # 'OntologyTags': ontology_tag_columns
                   }

    def set_properties(raw_df: pandas.DataFrame, column_dict: dict):
        for key, value in column_dict.items():

            def update_values(row):
                j_row = row[value].to_dict()
                return j_row

            if len(value) > 0:
                raw_df[key] = raw_df.apply(update_values, axis=1)
            else:
                raw_df[key] = None

        return raw_df

    data_frame = set_properties(raw_df=data_frame, column_dict=column_dict)
    data_frame = data_frame.drop(columns=discovery_properties_columns + ['Timestamp'] + device_tag_columns +
                                 sensor_tag_columns + metadata_columns  # + ontology_tag_columns
                                 )
    data_frame = data_frame.assign(OntologyTags=None)

    final_req_cols = ['ApiProjectId', 'InstallationId', 'NetworkDeviceId', 'DriverClassName', 'UserId', 'BatchNo',
                      'DriverUniqueId', 'CurrentValue', 'DriverDeviceType', 'ObjectPropertyTemplateName',
                      'UnitOfMeasureAbbrev', 'DeviceName', 'DisplayName', 'EquipmentType', 'EquipmentLabel',
                      'DeviceTags', 'SensorTags', 'OntologyTags', 'Metadata', 'DiscoveryProperties', 'JobId']

    if set(data_frame.columns.tolist()).issubset(final_req_cols):
        batch_size = 50
        chunk_list = []
        payload_error_list = []
        grouped_df = data_frame.reset_index(drop=True).groupby(by=lambda x: x // batch_size, axis=0)

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/integrations/driver-discovery"
        headers = api_inputs.api_headers.default

        logger.info(f"Input has been batched into {grouped_df.ngroups} group(s). ")

        for name, group in grouped_df:
            discovery_payload = group.groupby(
                by=['ApiProjectId', 'InstallationId', 'NetworkDeviceId', 'DriverClassName', 'UserId', 'BatchNo', 'JobId']).apply(
                lambda x: x[['DriverUniqueId', 'CurrentValue', 'DriverDeviceType', 'ObjectPropertyTemplateName',
                             'UnitOfMeasureAbbrev', 'DeviceName', 'DisplayName', 'EquipmentType', 'EquipmentLabel',
                             'DeviceTags', 'SensorTags', 'OntologyTags', 'Metadata', 'DiscoveryProperties']].to_dict(
                    orient='records')).reset_index().rename(columns={0: 'Sensors'}).to_json(orient='records')

            # remove outer [] from json
            discovery_payload = re.sub(r"^[\[]", '', re.sub(r"[\]]$", '', discovery_payload))

            logger.info(f"Upserting discovery record group {name} of {grouped_df.ngroups - 1}. ")
            response = requests.post(url=url, headers=headers, data=discovery_payload)

            response_status = '{} {}'.format(response.status_code, response.reason)
            if response.status_code != 200 and len(response.text) > 0:
                logger.error(f"API Call was not successful. Response Status: {response.status_code}. "
                             f"Reason: {response.reason}. Error Text: {response.text}")
                payload_error_list += [{'Chunk': name, 'Payload': discovery_payload}]
                if response.text.startswith('{'):
                    response_content = response.json()
                    chunk_list += [{'Chunk': name, 'response_status': response.status_code,
                                    'response_reason': response.reason, 'error_code': response_content['ErrorCode'],
                                    'errors': response_content['Errors']}]
                elif response.text.startswith('"'):
                    chunk_list += [{'Chunk': name, 'response_status': response.status_code,
                                    'response_reason': response.reason, 'errors': response.json()}]
                else:
                    chunk_list += [{'Chunk': name, 'response_status': response.status_code,
                                    'response_reason': response.reason, 'errors': response.text}]
            elif response.status_code != 200 and len(response.text) == 0:
                logger.error(f"API Call was not successful. Response Status: {response.status_code}. "
                             f"Reason: {response.reason}. ")
                payload_error_list += [{'Chunk': name, 'Payload': discovery_payload}]
                chunk_list += [{'Chunk': name, 'response_status': response.status_code,
                                'response_reason': response.reason}]
            elif response.status_code == 200:
                logger.info(f"API Call was successful. ")
                chunk_list += [{'Chunk': name, 'response_status': response.status_code,
                                'response_reason': response.reason}]

        upsert_response_df = pandas.DataFrame(chunk_list)

        if 0 < len(payload_error_list) <= grouped_df.ngroups:
            payload_error_df = pandas.DataFrame(payload_error_list)
            logger.error(f"Errors on upsert of discovered records. ")
            return upsert_response_df, payload_error_df

    return upsert_response_df, pandas.DataFrame()


@_with_func_attrs(df_required_columns=['InstallationCode', 'DeviceCode', 'DeviceName', 'SensorName', 'SensorTemplate',
                                       'SensorUnitOfMeasure', 'EquipmentClass', 'EquipmentLabel'])
def upsert_device_sensors(df: pandas.DataFrame, api_inputs: ApiInputs, tag_columns: list = None,
                          metadata_columns: list = None, save_additional_columns_as_slices: bool = False):
    """Upsert device(s) and sensor(s)

    Required fields are:

    - InstallationCode
    - DeviceCode
    - DeviceName
    - SensorName
    - SensorTemplate
    - SensorUnitOfMeasure
    - EquipmentClass
    - EquipmentLabel

    Parameters
    ----------
    df: pandas.DataFrame
        The asset register created by the driver including the minimum required set of columns.
    api_inputs : ApiInputs
        Object returned by initialize() function.
    tag_columns : list, default = None
        Columns of dataframe that contain tags (Default value = None).
    metadata_columns : list, default = None
        Column(s) of dataframe that contain device-level metadata (Default value = None).
    save_additional_columns_as_slices : bool, default = False
        Whether additional columns should be saved as slices (Default value = False).

    Returns
    -------
    tuple[list, pandas.DataFrame]
        (response_status_list, upsert_response_df) - Returns the list of response statuses and the dataframe containing
        the parsed response text.

    """

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    data_frame = df.copy()

    required_columns = ['InstallationCode', 'DeviceCode', 'DeviceName', 'SensorName', 'SensorTemplate',
                        'SensorUnitOfMeasure', 'EquipmentClass', 'EquipmentLabel']
    proposed_columns = data_frame.columns.tolist()

    if not set(required_columns).issubset(data_frame.columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'Integration.upsert_device_sensors(): data_frame must contain the following columns: ' + ', '.join(
            required_columns)

    if tag_columns is not None and not set(tag_columns).issubset(data_frame.columns):
        logger.exception('Missing expected tag column(s): %s', set(tag_columns).difference(proposed_columns))
        return 'Integration.upsert_device_sensors(): data_frame expected to contain the following tag column(s): ' + \
               ', '.join(tag_columns)
    elif tag_columns is None:
        tag_columns = []

    if metadata_columns is not None and not set(metadata_columns).issubset(data_frame.columns):
        logger.exception('Missing expected metadata column(s): %s', set(metadata_columns).difference(proposed_columns))
        return 'Integration.upsert_device_sensors(): data_frame expected to contain the following metadata ' \
               'column(s): ' + ', '.join(metadata_columns)
    elif metadata_columns is None:
        metadata_columns = []

    slice_columns = set(proposed_columns).difference(set(required_columns)) - set(tag_columns) - set(metadata_columns)
    slices_data_frame = pandas.DataFrame()

    if len(slice_columns) > 0 or len(tag_columns) > 0 or len(metadata_columns) > 0:
        def update_values(row, mode):
            if mode == 'A':
                j_row = row[slice_columns].to_json()
                return str(j_row)
            elif mode == 'B':
                j_row = row[tag_columns].to_json()
                return str(j_row)
            else:
                j_row = row[metadata_columns].to_json()
                return str(j_row)

        data_frame['Slices'] = data_frame.apply(update_values, args="A", axis=1)
        data_frame['TagsJson'] = data_frame.apply(update_values, args="B", axis=1)
        data_frame['MetadataJson'] = data_frame.apply(update_values, args="C", axis=1)
        data_frame = data_frame.drop(columns=slice_columns)
        slices_data_frame = data_frame[['DeviceCode', 'Slices']]

    headers = api_inputs.api_headers.integration

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/devices/upsert-ingestion"

    data_frame_grpd = data_frame.groupby(['InstallationCode', 'DeviceCode'])
    chunk_list = []
    for name, group in data_frame_grpd:
        logger.info("Sending request: POST %s", url)
        logger.info('Upserting data for InstallationCode = %s and DeviceCode = %s', str(name[0]), str(name[1]))
        # logger.info('Sensor count to upsert: %s', str(group.shape[0]))

        response = requests.post(url, data=group.to_json(orient='records'), headers=headers)

        response_status = '{} {}'.format(response.status_code, response.reason)
        logger.info("Response status: %s", response_status)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            chunk_list += [{'Chunk': name, 'DeviceCountToUpsert': 1, 'SensorCountToUpsert': str(group.shape[0]),
                            'response_status': response_status, 'response_df': pandas.DataFrame(),
                            'invalid_rows': pandas.DataFrame()}]
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            chunk_list += [{'Chunk': name, 'DeviceCountToUpsert': 1, 'SensorCountToUpsert': str(group.shape[0]),
                            'response_status': response_status, 'response_df': pandas.DataFrame(),
                            'invalid_rows': pandas.DataFrame()}]
        elif response.status_code == 200 and len(response.text) > 0:
            response_data_frame = pandas.read_json(response.text)
            logger.info('Dataframe response row count = %s', str(response_data_frame.shape[0]))
            if response_data_frame.shape[1] > 0:
                response_data_frame = response_data_frame.assign(InstallationCode=str(name[0]),
                                                                 SensorCount=group.shape[0])
                invalid_rows = response_data_frame[response_data_frame['status'] != 'Ok']
                if invalid_rows.shape[0] > 0:
                    logger.error("The following rows contain invalid data: %s", invalid_rows)
                    chunk_list += [
                        {'Chunk': name, 'DeviceCountToUpsert': 1,
                         'SensorCountToUpsert': str(group.shape[0]),
                         'response_status': response_status,
                         'response_df': response_data_frame[response_data_frame['status'] == 'Ok'],
                         'invalid_rows': invalid_rows}]
                else:
                    chunk_list += [{'Chunk': name, 'DeviceCountToUpsert': 1,
                                    'SensorCountToUpsert': str(group.shape[0]),
                                    'response_status': response_status, 'response_df': response_data_frame,
                                    'invalid_rows': invalid_rows}]

    upsert_response_df = pandas.DataFrame()
    upsert_invalid_rows_df = pandas.DataFrame()
    upsert_response_status_list = []
    for i in range(len(chunk_list)):
        upsert_response_df = upsert_response_df.append(chunk_list[i]['response_df'])
        upsert_invalid_rows_df = upsert_invalid_rows_df.append(chunk_list[i]['invalid_rows'])
        upsert_response_status_list += [chunk_list[i]['response_status']]

    if save_additional_columns_as_slices and slices_data_frame.shape[0] > 0:
        slices_merged = pandas.merge(left=upsert_response_df, right=slices_data_frame, left_on='deviceCode',
                                     right_on='DeviceCode')
        slices_data_frame = slices_merged[['deviceId', 'Slices']]
        slices_data_frame = slices_data_frame.rename(columns={'deviceId': 'DeviceId'})
        upsert_data(slices_data_frame, api_inputs=api_inputs, key_columns=['DeviceId'], table_name='DeviceSensorSlices',
                    is_slices_table=True)

    upsert_response_df.columns = _column_name_cap(columns=upsert_response_df.columns)
    _upsert_entities_affected_count(api_inputs=api_inputs,
                                    entities_affected_count=upsert_response_df['SensorCount'].sum())
    _adx_support(api_inputs=api_inputs, payload_type='Sensors')

    logger.info("Ingestion Complete. ")
    return upsert_response_status_list, upsert_response_df


@_with_func_attrs(df_required_columns=['InstallationName', 'InstallationCode', 'Address', 'Country', 'Suburb', 'State',
                                       'StateName', 'FloorAreaM2', 'ZipPostCode'],
                  df_optional_columns=['Latitude', 'Longitude', 'Timezone', 'InstallationId'])
def upsert_sites(df: pandas.DataFrame, api_inputs: ApiInputs, tag_columns: list = None,
                 save_additional_columns_as_slices: bool = False):
    """Upsert site(s).

    The `df` input must contain the following columns:
        - InstallationName
        - InstallationCode
        - Address
        - Suburb
        - State
        - StateName
        - Country
        - FloorAreaM2
        - ZipPostCode

    The following additional columns are optional:
        - Latitude
        - Longitude
        - Timezone
        - InstallationId
            - The UUID of the existing site within the Switch Automation Platform.

    Parameters
    ----------
    df: pandas.DataFrame :
        The dataframe containing the sites to be created/updated in the Switch platform. All required columns must be
        present with no null values.
    api_inputs : ApiInputs
        Object returned by initialize() function.
    tag_columns : list, default=[]
        The columns containing site-level tags. The column header will be the tag group name. (Default value = True)
    save_additional_columns_as_slices : bool, default = False
        Whether any additional columns should be saved as slices. (Default value = False)

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response, response_data_frame) - Returns the response status and the dataframe containing the parsed response
        text.

    """

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    data_frame = df.copy()

    required_columns = ['InstallationName', 'InstallationCode', 'Address', 'Country', 'Suburb', 'State', 'StateName',
                        'FloorAreaM2', 'ZipPostCode']
    optional_columns = ['Latitude', 'Longitude', 'Timezone', 'InstallationId']
    proposed_columns = data_frame.columns.tolist()

    if not set(required_columns).issubset(proposed_columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'Integration.upsert_sites() - data_frame must contain the following columns: ' + ', '.join(
            required_columns) + '. Optional Columns include: ' + ', '.join(optional_columns)

    if tag_columns is not None and not set(tag_columns).issubset(data_frame.columns):
        logger.exception('Missing expected tag column(s): %s', set(tag_columns).difference(proposed_columns))
        return 'Integration.upsert_sites(): data_frame must contain the following tag columns: ' + ', '.join(
            tag_columns)
    elif tag_columns is None:
        tag_columns = []

    slice_columns = set(proposed_columns).difference(set(required_columns + optional_columns)) - set(tag_columns)
    slices_data_frame = pandas.DataFrame()

    if len(slice_columns) > 0 or len(tag_columns) > 0:
        def update_values(row, mode):
            if mode == 'A':
                j_row = row[slice_columns].to_json()
                return str(j_row)
            else:
                j_row = row[tag_columns].to_json()
                return str(j_row)

        data_frame['Slices'] = data_frame.apply(update_values, args=('A',), axis=1)
        data_frame['TagsJson'] = data_frame.apply(update_values, args=('B',), axis=1)

        data_frame = data_frame.drop(columns=slice_columns)
        slices_data_frame = data_frame[['InstallationCode', 'Slices']]

    headers = api_inputs.api_headers.integration

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/installations/upsert-ingestion"
    logger.info("Sending request: POST %s", url)

    response = requests.post(url, data=data_frame.to_json(orient='records'), headers=headers)

    response_data_frame = pandas.read_json(response.text)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    if save_additional_columns_as_slices and slices_data_frame.shape[0] > 0:
        slices_merged = pandas.merge(left=response_data_frame, right=slices_data_frame, left_on='installationCode',
                                     right_on='InstallationCode')
        slices_data_frame = slices_merged[['installationId', 'Slices']]
        slices_data_frame = slices_data_frame.rename(columns={'installationId': 'InstallationId'})
        upsert_data(slices_data_frame, api_inputs=api_inputs, key_columns=['InstallationId'],
                    table_name='InstallationSlices', is_slices_table=True)

    response_data_frame.columns = _column_name_cap(columns=response_data_frame.columns)

    count_entities = response_data_frame.apply(
        lambda row: 'Created' if (row['IsInserted'] == True and row['IsUpdated'] == False) else (
            'Updated' if row['IsInserted'] == False and row['IsUpdated'] == True else 'Failed'), axis=1).isin(
        ['Created', 'Updated']).sum()

    _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=count_entities)
    _adx_support(api_inputs=api_inputs, payload_type='Sites')

    logger.info("Ingestion complete. ")

    return response_status, response_data_frame


@_with_func_attrs(df_required_columns=['WorkOrderId', 'InstallationId', 'WorkOrderSiteIdentifier', 'Status',
                                       'RawStatus', 'Priority', 'RawPriority', 'WorkOrderCategory',
                                       'RawWorkOrderCategory', 'Type', 'Description', 'CreatedDate',
                                       'LastModifiedDate', 'WorkStartedDate', 'WorkCompletedDate', 'ClosedDate'],
                  df_optional_columns=['SubType', 'Vendor', 'VendorId', 'EquipmentClass', 'RawEquipmentClass',
                                       'EquipmentLabel', 'RawEquipmentId', 'TenantId', 'TenantName', 'NotToExceedCost',
                                       'TotalCost', 'BillableCost', 'NonBillableCost', 'Location', 'RawLocation',
                                       'ScheduledStartDate', 'ScheduledCompletionDate'])
def upsert_workorders(df: pandas.DataFrame, api_inputs: ApiInputs, save_additional_columns_as_slices: bool = False):
    """Upsert data to the Workorder table.

    The following columns are required to be present in the df:

    - ``WorkOrderId``: unique identifier for the work order instance
    - ``InstallationId``: the InstallationId (guid) used to uniquely identify a given site within the Switch platform
    - ``WorkOrderSiteIdentifier``: the work order provider's raw/native site identifier field
    - ``Status``: the status mapped to the Switch standard values defined by literal: `WORK_ORDER_STATUS`
    - ``RawStatus``: the work order provider's raw/native status
    - ``Priority``: the priority mapped to the Switch standard values defined by literal: `WORK_ORDER_PRIORITY`
    - ``RawPriority``: the work order provider's raw/native priority
    - ``WorkOrderCategory``: the category mapped to the Switch standard values defined by literal: `WORK_ORDER_CATEGORY`
    - ``RawWorkOrderCategory``: the work order provider's raw/native category
    - ``Type`` - work order type (as defined by provider) - e.g. HVAC - Too Hot, etc.
    - ``Description``: description of the work order.
    - ``CreatedDate``: the date the work order was created (Submitted status)
    - ``LastModifiedDate``: datetime the workorder was last modified
    - ``WorkStartedDate``: datetime work started on the work order (In Progress status)
    - ``WorkCompletedDate``: datetime work was completed for the work order (Resolved status)
    - ``ClosedDate``: datetime the workorder was closed (Closed status)

    The following columns are optional:

    - ``SubType``: the sub-type of the work order
    - ``Vendor``: the name of the vendor
    - ``VendorId``: the vendor id
    - ``EquipmentClass``: the Switch defined Equipment Class mapped from the work order provider's definition
    - ``RawEquipmentClass``: the work order provider's raw/native equipment class
    - ``EquipmentLabel``: the EquipmentLabel as defined within the Switch platform
    - ``RawEquipmentId``: the work order provider's raw/native equipment identifier/label
    - ``TenantId``: the tenant id
    - ``TenantName``: the name of the tenant
    - ``NotToExceedCost``: the cost not to be exceeded for the given work order
    - ``TotalCost``: total cost of the work order
    - ``BillableCost``: the billable portion of the work order cost
    - ``NonBillableCost``: the non-billable portion of the work order cost.
    - ``Location``: the Location as defined within the Switch platform
    - ``RawLocation``: the work order provider's raw/native location definition
    - ``ScheduledStartDate``: datetime work was scheduled to start on the given work order
    - ``ScheduledCompletionDate``" datetime work was scheduled to be completed for the given work order


    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe containing the work order data to be upserted.
    api_inputs: ApiInputs
        Object returned by initialize() function.
    save_additional_columns_as_slices : bool, default = False
         (Default value = False)

    Returns
    -------

    """

    data_frame = df.copy()

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    required_columns = ['WorkOrderId', 'InstallationId', 'WorkOrderSiteIdentifier', 'Status',
                        'RawStatus', 'Priority', 'RawPriority', 'WorkOrderCategory', 'RawWorkOrderCategory', 'Type',
                        'Description', 'CreatedDate', 'LastModifiedDate', 'WorkStartedDate', 'WorkCompletedDate',
                        'ClosedDate']
    optional_columns = ['SubType', 'Vendor', 'VendorId', 'EquipmentClass', 'RawEquipmentClass', 'EquipmentLabel',
                        'RawEquipmentId', 'TenantId', 'TenantName', 'NotToExceedCost', 'TotalCost', 'BillableCost',
                        'NonBillableCost', 'Location', 'RawLocation', 'ScheduledStartDate', 'ScheduledCompletionDate']

    req_cols = ', '.join(required_columns)
    proposed_columns = data_frame.columns.tolist()

    if not set(required_columns).issubset(proposed_columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'Integration.upsert_workorder() - data_frame must contain the following columns: ' + req_cols

    try:
        _work_order_schema.validate(data_frame, lazy=True)
    except pandera.errors.SchemaErrors as err:
        logger.error('Errors present with columns in df provided.')
        logger.error(err.failure_cases)
        schema_error = err.failure_cases
        return schema_error

    slice_columns = set(proposed_columns).difference(set(required_columns + optional_columns))

    missing_optional_columns = set(optional_columns) - set(proposed_columns)
    for missing_column in missing_optional_columns:
        data_frame[missing_column] = ''

    if len(slice_columns) > 0 and save_additional_columns_as_slices is True:
        def update_values(row):
            j_row = row[slice_columns].to_json()
            return str(j_row)

        data_frame['Meta'] = data_frame.apply(update_values, axis=1)
        data_frame = data_frame.drop(columns=slice_columns)
    elif len(slice_columns) > 0 and save_additional_columns_as_slices is not True:
        data_frame = data_frame.drop(columns=slice_columns)
        data_frame['Meta'] = ''
    else:
        data_frame['Meta'] = ''

    # payload = {}
    headers = api_inputs.api_headers.integration

    logger.info("Upserting data to Workorders.")

    data_frame = data_frame.loc[:, ['WorkOrderId', 'InstallationId', 'WorkOrderSiteIdentifier', 'Status', 'Priority',
                                    'WorkOrderCategory', 'Type', 'Description', 'CreatedDate', 'LastModifiedDate',
                                    'WorkStartedDate', 'WorkCompletedDate', 'ClosedDate', 'RawPriority',
                                    'RawWorkOrderCategory', 'RawStatus', 'SubType', 'Vendor', 'VendorId',
                                    'EquipmentClass', 'RawEquipmentClass', 'EquipmentLabel', 'RawEquipmentId',
                                    'TenantId', 'TenantName', 'NotToExceedCost', 'TotalCost', 'BillableCost',
                                    'NonBillableCost', 'Location', 'RawLocation', 'ScheduledStartDate',
                                    'ScheduledCompletionDate', 'Meta']]

    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name='WorkOrder')
    json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "append",
                    "tableDef": _get_structure(data_frame),
                    "keyColumns": ["WorkOrderId"]
                    }

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/adx/work-order-operation"
    logger.info("Sending request: POST %s", url)

    logger.info("Sending request to ingest %s files from %s", str(upload_result[1]), upload_result[0])
    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    elif response.status_code == 200:
        _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=data_frame.shape[0])

    response_df = pandas.read_json(response.text, typ='series').to_frame().T
    response_df.columns = _column_name_cap(columns=response_df.columns)

    logger.info("Response status: %s", response_status)
    logger.info("Ingestion complete. ")

    return response_status, response_df


@_with_func_attrs(df_required_columns=['ObjectPropertyId', 'InstallationId', 'Timestamp', 'Value'])
def upsert_timeseries_ds(df: pandas.DataFrame, api_inputs: ApiInputs, is_local_time: bool = True,
                         save_additional_columns_as_slices: bool = False, data_feed_file_status_id: uuid.UUID = None):
    """Upserts to Timeseries_Ds table.

    The following columns are required to be present in the data_frame:
    - InstallationId
    - ObjectPropertyId
    - Timestamp
    - Value

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe containing the data to be appended to timeseries.
    api_inputs: ApiInputs
        Object returned by initialize() function.
    is_local_time : bool, default = True
         Whether the datetime values are in local time or UTC. If false, then UTC (Default value = True).
    save_additional_columns_as_slices : bool, default = False
         (Default value = False)
    data_feed_file_status_id : uuid.UUID, default = None
         Enables developer to identify upserted rows using during development. This data is posted to the
         DataFeedFileStatusId in the Timeseries_Ds table.

         Once deployed, the DataFeedFileStatusId field will contain a unique Guid which will assist in
         tracking upload results and logging.

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response_status, response_df) - Returns the response status and the dataframe containing the parsed response
        text.

    """

    data_frame = df.copy()

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    required_columns = ['ObjectPropertyId', 'InstallationId', 'Timestamp', 'Value']
    req_cols = ', '.join(required_columns)
    proposed_columns = data_frame.columns.tolist()

    if not set(required_columns).issubset(proposed_columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'integration.upsert_timeseries_ds() - data_frame must contain the following columns: ' + req_cols

    slice_columns = set(proposed_columns).difference(set(required_columns))

    if len(slice_columns) > 0 and save_additional_columns_as_slices is True:
        def update_values(row):
            j_row = row[slice_columns].to_json()
            return str(j_row)

        # data_frame['Meta'] = data_frame.apply(update_values, axis=1)
        data_frame['Meta'] = data_frame[slice_columns].assign(
            **data_frame[slice_columns].select_dtypes(['datetime', 'object']).astype(str)).apply(update_values, axis=1)

        data_frame = data_frame.drop(columns=slice_columns)
    elif len(slice_columns) > 0 and save_additional_columns_as_slices is not True:
        data_frame = data_frame.drop(columns=slice_columns)
        data_frame['Meta'] = ''
    else:
        data_frame['Meta'] = ''

    if api_inputs.data_feed_file_status_id is not None and api_inputs.data_feed_file_status_id != '00000000-0000-0000' \
                                                                                                  '-0000-000000000000':
        data_frame['DataFeedFileStatusId'] = api_inputs.data_feed_file_status_id
    elif data_feed_file_status_id is not None:
        data_frame['DataFeedFileStatusId'] = data_feed_file_status_id
    else:
        data_frame['DataFeedFileStatusId'] = '00000000-0000-0000-0000-000000000000'

    site_list = data_frame['InstallationId'].unique().tolist()
    start_date = data_frame['Timestamp'].min(axis=0, skipna=True)
    end_date = data_frame['Timestamp'].max(axis=0, skipna=True)

    timezones = _timezone_offsets(api_inputs=api_inputs, date_from=start_date.date(), date_to=end_date.date(),
                                  installation_id_list=site_list)
    timezones['dateFrom'] = timezones['dateFrom'].apply(lambda x: pandas.to_datetime(x))
    timezones['dateTo'] = timezones['dateTo'].apply(lambda x: pandas.to_datetime(x))

    data_frame = data_frame.merge(timezones, left_on='InstallationId', right_on='installationId', how='inner')

    def in_range(row):
        j_row = (row['Timestamp'] >= row['dateFrom']) & (
            row['Timestamp'] < (row['dateTo'] + datetime.timedelta(days=1)))
        return str(j_row)

    data_frame['InDateRange'] = data_frame.apply(in_range, axis=1)
    data_frame = data_frame[data_frame['InDateRange'] == 'True']

    if is_local_time:
        def to_utc(row):
            if row['offsetToUtcMinutes'] >= 0:
                j_row = row['TimestampLocal'] - datetime.timedelta(minutes=row['offsetToUtcMinutes'])
            else:
                j_row = row['TimestampLocal'] + datetime.timedelta(minutes=abs(row['offsetToUtcMinutes']))
            return j_row

        data_frame = data_frame.assign(TimestampLocal=data_frame['Timestamp'])
        data_frame['Timestamp'] = data_frame.apply(to_utc, axis=1)
    elif not is_local_time:
        def from_utc(row):
            if row['offsetToUtcMinutes'] >= 0:
                j_row = row['Timestamp'] + datetime.timedelta(minutes=row['offsetToUtcMinutes'])
            else:
                j_row = row['Timestamp'] - datetime.timedelta(minutes=abs(row['offsetToUtcMinutes']))
            return j_row

        data_frame['TimestampLocal'] = data_frame.apply(from_utc, axis=1)

    def bin_to_15_minute_interval(row, date_col):
        if row[date_col].minute < 15:
            j_row = row[date_col].replace(minute=0)
            return j_row
        elif row[date_col].minute >= 15 and row[date_col].minute < 30:
            j_row = row[date_col].replace(minute=15)
            return j_row
        elif row[date_col].minute >= 30 and row[date_col].minute < 45:
            j_row = row[date_col].replace(minute=30)
            return j_row
        else:
            j_row = row[date_col].replace(minute=45)
            return j_row

    data_frame['TimestampId'] = data_frame.apply(bin_to_15_minute_interval, args=('Timestamp',), axis=1)
    data_frame['TimestampLocalId'] = data_frame.apply(bin_to_15_minute_interval, args=('TimestampLocal',), axis=1)
    data_frame = data_frame.drop(columns=['InDateRange', 'dateFrom', 'dateTo', 'installationId', 'offsetToUtcMinutes'])

    # payload = {}
    headers = api_inputs.api_headers.integration

    logger.info("Upserting data to Timeseries_Ds.")

    data_frame = data_frame.loc[:, ['ObjectPropertyId', 'Timestamp', 'TimestampId', 'TimestampLocal',
                                    'TimestampLocalId', 'Value', 'DataFeedFileStatusId', 'InstallationId', 'Meta']]
    name = 'Timeseries_Ds'

    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=name, batch_id=data_feed_file_status_id)

    json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "upsert",
                    "isLocalTime": is_local_time,
                    "tableDef": _get_structure(data_frame),
                    "keyColumns": ["ObjectPropertyId", "Timestamp"]
                    }

    # {'ObjectPropertyId': 'object', 'Timestamp': 'datetime64[ns]', 'Value': 'float64', 'InstallationId': 'object'}

    # ObjectPropertyId: string, Timestamp: datetime, TimestampId: datetime, TimestampLocal: datetime,
    # TimestampLocalId: datetime, Value: real, DataFeedFileStatusId: string, InstallationId: string, Meta: dynamic

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/adx/time-series-operation?originalName=" \
          f"{name}"
    logger.info("Sending request: POST %s", url)

    logger.info("Sending request to ingest %s files from %s", str(upload_result[1]), upload_result[0])
    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    elif response.status_code == 200:
        _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=data_frame.shape[0])

    response_df = pandas.read_json(response.text, typ='series').to_frame().T
    response_df.columns = _column_name_cap(columns=response_df.columns)

    logger.info("Response status: %s", response_status)
    logger.info("Ingestion complete. ")
    return response_status, response_df


@_with_func_attrs(df_required_columns=['ObjectPropertyId', 'InstallationId', 'Timestamp', 'Value'])
def upsert_timeseries(df: pandas.DataFrame, api_inputs: ApiInputs, is_local_time: bool = True,
                      save_additional_columns_as_slices: bool = False, data_feed_file_status_id: uuid.UUID = None, 
                      is_specific_timezone: Union[bool, str] = False, ingestion_mode: IngestionMode = 'Queue'):
    """Upserts timeseries to EventHub for processing.

    The following columns are required to be present in the data_frame:
    - InstallationId
    - ObjectPropertyId
    - Timestamp
    - Value

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe containing the data to be appended to timeseries.
    api_inputs: ApiInputs
        Object returned by initialize() function.
    is_local_time : bool, default = True
         Whether the datetime values are in local time or UTC. If, False and is_specific_timezone is False, then UTC (Default value = True).
         Should be set to False when 'is_specific_timezone' has value.
    save_additional_columns_as_slices : bool, default = False
         (Default value = False)
    data_feed_file_status_id : uuid.UUID, default = None
         Enables developer to identify upserted rows using during development. This data is posted to the
         DataFeedFileStatusId in the Timeseries_Ds table.

         Once deployed, the DataFeedFileStatusId field will contain a unique Guid which will assist in
         tracking upload results and logging.
    is_specific_timezone : Union[False, str]
        Accepts a timezone name as the specific timezone used by the source data. Defaults to False.
        Cannot have value if 'is_local_time' is set to True.
        Retrieve list of timezones using 'sw.integration.get_timezones()'

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response_status, response_df) - Returns the response status and the dataframe containing the parsed response
        text.

    """

    data_frame = df.copy()
    timezones_df = pd.DataFrame()
    logger.info(f'Data frame size: {convert_bytes(data_frame.memory_usage(deep=True).sum())}')

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if not isinstance(is_specific_timezone, str) and not isinstance(is_specific_timezone, bool):
        logger.error("'is_specific_timezone' parameter can only be of type str or bool.")
        return pandas.DataFrame()

    if is_specific_timezone != False and is_local_time == True:
        logger.error("Assigning specific timezone is only possible if 'is_local_time' is set to False.")
        return pandas.DataFrame()

    if is_specific_timezone == True:
        logger.error("'is_specific_timezone' parameter value, if not False, should be a valid timezone. Retrieve list of timezones using 'sw.integration.get_timezones()")
        return pandas.DataFrame()

    if is_specific_timezone == '':
        logger.error("'is_specific_timezone' cannot be an empty string. Retrieve list of timezones using 'sw.integration.get_timezones()")
        return pandas.DataFrame()

    is_specific_timezone_conversion = False
    if is_specific_timezone != False and (isinstance(is_specific_timezone, str) and is_specific_timezone != ''):
        timezone_name_df = get_timezones(api_inputs=api_inputs)

        if timezone_name_df is None or timezone_name_df.__len__() == 0:
            logger.error("Failed to retrieve the allowed Timezone names.")
            return pandas.DataFrame()
        
        if not is_specific_timezone in timezone_name_df.TimezoneName.values:
            logger.error(f"'is_specific_timezone' parameter value '{is_specific_timezone}' is not found in Timezone Names list.")
            logger.info(f"'Timezone name list can be retrieved using 'sw.integration.get_timezones()' method.")
            return pandas.DataFrame()
        
        is_specific_timezone_conversion = True

    required_columns = ['ObjectPropertyId', 'InstallationId', 'Timestamp', 'Value']
    req_cols = ', '.join(required_columns)
    proposed_columns = data_frame.columns.tolist()

    if not set(required_columns).issubset(proposed_columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'integration.upsert_timeseries() - data_frame must contain the following columns: ' + req_cols

    if not pandas.api.types.is_datetime64_ns_dtype(data_frame['Timestamp']):
        logger.exception("Timestamp series value is not of type: datetime64[ns] dtype.")
        return "Timestamp series value is not of type: datetime64[ns] dtype."

    null_count_timestamp = data_frame['Timestamp'].isna().sum()
    if null_count_timestamp > 0:
        logger.warning(f"There are {null_count_timestamp} records with a null or empty Timestamp value. " \
            f"These are being dropped resulting in {data_frame.shape[0] - null_count_timestamp} records will be upserted.")
        data_frame = data_frame.dropna(subset=['Timestamp'])

    slice_columns = set(proposed_columns).difference(set(required_columns))

    if len(list(slice_columns)) > 0 and save_additional_columns_as_slices is True:
        def update_values(row):
            j_row = row[list(slice_columns)].to_json()
            return str(j_row)

        logger.info('Creating Meta Column.')
        data_frame['Meta'] = data_frame[list(slice_columns)].assign(
            **data_frame[list(slice_columns)].select_dtypes(['datetime', 'object']).astype(str)).apply(update_values, axis=1)

        data_frame = data_frame.drop(columns=list(slice_columns))
    elif len(list(slice_columns)) > 0 and save_additional_columns_as_slices is not True:
        data_frame = data_frame.drop(columns=list(slice_columns))
        data_frame['Meta'] = ''
    else:
        data_frame['Meta'] = ''

    if api_inputs.data_feed_file_status_id is not None and \
            api_inputs.data_feed_file_status_id != '00000000-0000-0000-0000-000000000000' and \
            api_inputs.data_feed_file_status_id != '':
        data_frame['DataFeedFileStatusId'] = api_inputs.data_feed_file_status_id
        data_feed_file_status_id = api_inputs.data_feed_file_status_id
    elif data_feed_file_status_id is not None and \
            data_feed_file_status_id != '00000000-0000-0000-0000-000000000000' and \
            data_feed_file_status_id != '':
        data_frame['DataFeedFileStatusId'] = data_feed_file_status_id
    else:
        data_feed_file_status_id = uuid.uuid4()
        data_frame['DataFeedFileStatusId'] = data_feed_file_status_id

    
    start_date = data_frame['Timestamp'].min(axis=0, skipna=True)
    end_date = data_frame['Timestamp'].max(axis=0, skipna=True)

    specific_timezone_df = pd.DataFrame()

    def convert_dst_interval_dates(row):
        row['start'] = pd.to_datetime(row['start'])
        row['end'] = pd.to_datetime(row['end'])
        row['offsetToUtcMinutes'] = row['standardOffsetUtc'] + row['dstOffsetUtc']
        return row

    if is_specific_timezone_conversion:
        def timezone_to_utc(row):
            if row['offsetToUtcMinutes'] >= 0:
                j_row = row['Timestamp'] - datetime.timedelta(minutes=row['offsetToUtcMinutes'])
            else:
                j_row = row['Timestamp'] + datetime.timedelta(minutes=abs(row['offsetToUtcMinutes']))
            return j_row

        specific_timezone_df = _timezone_dst_offsets(api_inputs=api_inputs, date_from=start_date.date(), date_to=end_date.date(), timezone_name=is_specific_timezone)

        if specific_timezone_df.empty:
            logger.exception('Failed to retrieve the specific timezone offsets.')
            return 'Failed to retrieve the specific timezone offsets.'

        specific_timezone_df = specific_timezone_df.apply(convert_dst_interval_dates, axis=1)
        specific_timezone_df = specific_timezone_df.drop(columns=['timezoneId', 'standardOffsetUtc', 'dstOffsetUtc'])
        
        data_frame['year'] = pd.DatetimeIndex(data_frame['Timestamp']).year
        data_frame = data_frame.merge(specific_timezone_df, left_on='year', right_on='year', how='inner')

        del specific_timezone_df
        gc.collect()

        data_frame = data_frame[(data_frame.Timestamp >= data_frame.start) & (data_frame.Timestamp < data_frame.end) & (data_frame.Timestamp.apply(lambda x: x.year) == data_frame.year)]
        data_frame = data_frame.drop(columns=['start', 'end', 'year'])
        
        data_frame['Timestamp'] = data_frame.apply(timezone_to_utc, axis=1)
        data_frame = data_frame.drop(columns=['offsetToUtcMinutes'])

        # Set automatically to is_local_time = False 
        is_local_time = False

    site_list = data_frame['InstallationId'].unique().tolist()
    timezones_df = _timezone_dst_offsets(api_inputs=api_inputs, date_from=start_date.date(), date_to=end_date.date(),
                                         installation_id_list=site_list)

    if timezones_df.empty:
        logger.exception('Timezone DST offsets failed to retrieve.')
        return 'Timezone DST offsets failed to retrieve.'

    timezones_df = timezones_df.apply(convert_dst_interval_dates, axis=1)

    if not is_local_time:
        def convert_dst_interval_range_to_utc(row):
            dst_start = row['start']
            dst_end = row['end']

            if row['standardOffsetUtc'] >= 0:
                dst_start = dst_start - (datetime.timedelta(minutes=row['standardOffsetUtc'])) - datetime.timedelta(minutes=row['dstOffsetUtc'])
                dst_end = dst_end - (datetime.timedelta(minutes=row['standardOffsetUtc'])) - datetime.timedelta(minutes=row['dstOffsetUtc'])
            else:
                dst_start = dst_start + (datetime.timedelta(minutes=row['standardOffsetUtc'])) + datetime.timedelta(minutes=row['dstOffsetUtc'])
                dst_end = dst_end + (datetime.timedelta(minutes=row['standardOffsetUtc'])) + datetime.timedelta(minutes=row['dstOffsetUtc'])
            row['start'] = dst_start
            row['end'] = dst_end

            return row

        timezones_df = timezones_df.apply(convert_dst_interval_range_to_utc, axis=1)
    timezones_df = timezones_df.drop(columns=['timezoneId', 'standardOffsetUtc', 'dstOffsetUtc'])

    logger.info('Merging data frame input with timezone installation dst offsets.')
    data_frame = data_frame.merge(timezones_df, left_on='InstallationId', right_on='installationId', how='inner')
    
    del timezones_df
    gc.collect()

    data_frame = data_frame[(data_frame.Timestamp >= data_frame.start) & (data_frame.Timestamp < data_frame.end) & (data_frame.Timestamp.apply(lambda x: x.year) == data_frame.year)]
    data_frame = data_frame.drop(columns=['start', 'end', 'year'])

    if is_local_time:
        def to_utc(row):
            if row['offsetToUtcMinutes'] >= 0:
                j_row = row['TimestampLocal'] - datetime.timedelta(minutes=row['offsetToUtcMinutes'])
            else:
                j_row = row['TimestampLocal'] + datetime.timedelta(minutes=abs(row['offsetToUtcMinutes']))
            return j_row

        data_frame = data_frame.assign(TimestampLocal=data_frame['Timestamp'])
        data_frame['Timestamp'] = data_frame.apply(to_utc, axis=1)

    elif not is_local_time:
        def from_utc(row):
            if row['offsetToUtcMinutes'] >= 0:
                j_row = row['Timestamp'] + datetime.timedelta(minutes=row['offsetToUtcMinutes'])
            else:
                j_row = row['Timestamp'] - datetime.timedelta(minutes=abs(row['offsetToUtcMinutes']))
            return j_row
        data_frame['TimestampLocal'] = data_frame.apply(from_utc, axis=1)

    def bin_to_15_minute_interval(row, date_col):
        if row[date_col].minute < 15:
            j_row = row[date_col].replace(minute=0)
            return j_row
        elif row[date_col].minute >= 15 and row[date_col].minute < 30:
            j_row = row[date_col].replace(minute=15)
            return j_row
        elif row[date_col].minute >= 30 and row[date_col].minute < 45:
            j_row = row[date_col].replace(minute=30)
            return j_row
        else:
            j_row = row[date_col].replace(minute=45)
            return j_row

    logger.info('Date time conversion and formatting.')
    data_frame['TimestampId'] = data_frame.apply(bin_to_15_minute_interval, args=('Timestamp',), axis=1)
    data_frame['TimestampLocalId'] = data_frame.apply(bin_to_15_minute_interval, args=('TimestampLocal',), axis=1)

    timestamp_format = "%Y-%m-%dT%H:%M:%SZ"
    timestamp_normalized_format = "%Y-%m-%dT%H:%M:00Z"
    data_frame['Timestamp'] = data_frame['Timestamp'].dt.strftime(timestamp_format)
    data_frame['TimestampLocal'] = data_frame['TimestampLocal'].dt.strftime(timestamp_format)
    data_frame['TimestampId'] = data_frame['TimestampId'].dt.strftime(timestamp_normalized_format)
    data_frame['TimestampLocalId'] = data_frame['TimestampLocalId'].dt.strftime(timestamp_normalized_format)

    headers = api_inputs.api_headers.integration

    logger.info("Upserting data to Timeseries.")
    data_frame = data_frame.loc[:, ['ObjectPropertyId', 'Timestamp', 'TimestampId', 'TimestampLocal',
                                    'TimestampLocalId', 'Value', 'DataFeedFileStatusId', 'InstallationId', 'Meta']]

    container = 'data-ingestion-timeseries-adx'
    folder = 'to-adx-stream' if ingestion_mode == 'Stream' else 'to-eventhub'
    name = 'Timeseries'
    
    logger.info(f'Data frame size: {convert_bytes(data_frame.memory_usage(deep=True).sum())}')
    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, container=container, folder=folder, name=name, batch_id=data_feed_file_status_id, include_header=True)

    json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "ingestionMode": ingestion_mode}

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/timeseries"
    logger.info("Sending request: POST %s", url)

    logger.info("Sending request to ingest %s files from %s", str(upload_result[1]), upload_result[0])

    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)

    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    elif response.status_code == 200:
        _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=data_frame.shape[0])

    response_df = pandas.read_json(response.text, typ='series').to_frame().T
    response_df.columns = _column_name_cap(columns=response_df.columns)

    logger.info("Response status: %s", response_status)
    logger.info("Ingestion complete. ")
    return response_status, response_df


def upsert_data(data_frame, api_inputs: ApiInputs, table_name: str, key_columns: list,
                is_slices_table=False, ingestion_mode: IngestionMode = 'Queue'):
    """Upsert data

    Upserts data to the `table_name` provided to the function call and uses the `key_columns` provided to determine the
    unique records.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        Dataframe containing the data to be upserted.
    api_inputs : ApiInputs
        Object returned by initialize() function.
    table_name : str
        The name of the table where data will be upserted.
    key_columns : list
        The columns that determine a unique instance of a record. These are used to update records if new data is
        provided.
    is_slices_table : bool
         (Default value = False)

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response_status, response_df) - Returns the response status and the dataframe containing the parsed response
        text.

    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.integration

    if len(key_columns) == 0 or key_columns is None:
        logger.error(
            "You must provide key_columns. This allows the Switch Automation Platform to identify the rows to update.")
        return False

    logger.info("Data is being upserted for %s", table_name)

    table_structure = _get_structure(data_frame)
    if is_slices_table is True and 'Slices' in table_structure:
        table_structure['Slices'] = 'dynamic'

    # upload Blobs to folder
    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name)
    json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "upsert",
                    "tableDef": table_structure, "keyColumns": key_columns, 'ingestionMode': ingestion_mode}

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/adx/data-operation?tableName={table_name}"
    logger.info("Sending request: POST %s", url)

    logger.info("Sending request to ingest %s files from %s", str(upload_result[1]), upload_result[0])

    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    elif response.status_code == 200:
        _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=data_frame.shape[0])

    response_df = pandas.read_json(response.text, typ='series').to_frame().T
    response_df.columns = _column_name_cap(columns=response_df.columns)

    logger.info("Response status: %s", response_status)
    logger.info("Ingestion complete. ")

    return response_status, response_df


def replace_data(data_frame, api_inputs: ApiInputs, table_name: str):
    """Replace data

    Replaces the data in the ``table_name`` provided to the function call.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        Data frame to be used to replace the data in the `table_name` table.
    api_inputs : ApiInputs
        Object returned by initialize() function.
    table_name :
        The name of the table where data will be replaced.

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response_status, response_df) - Returns the response status and the dataframe containing the parsed response
        text.

    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.integration

    logger.info("Replacing all data for %s", table_name)

    # upload Blobs to folder
    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name)
    json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "replace",
                    "tableDef": _get_structure(data_frame)}
    logger.info(json_payload)

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/adx/data-operation?tableName={table_name}"
    logger.info("Sending request: POST %s", url)

    logger.info("Sending request to ingest %s files from %s", str(upload_result[1]), upload_result[0])

    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    elif response.status_code == 200:
        _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=data_frame.shape[0])

    response_df = pandas.read_json(response.text, typ='series').to_frame().T
    response_df.columns = _column_name_cap(columns=response_df.columns)

    logger.info("Ingestion complete. ")

    return response_status, response_df


def append_data(data_frame, api_inputs: ApiInputs, table_name: str):
    """Append data.

    Appends data to the ``table_name`` provided to the function call.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        Data to be appended.
    api_inputs : ApiInputs
        Object returned by initialize() function.
    table_name : str
        The name of the table where data will be appended.

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response_status, response_df) - Returns the response status and the dataframe containing the parsed response
        text.

    """
    # payload = {}
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    # payload = {}
    headers = api_inputs.api_headers.integration

    logger.info("Appending data for %s", table_name)

    # upload Blobs to folder
    upload_result = Blob.upload(api_inputs=api_inputs, data_frame=data_frame, name=table_name)
    json_payload = {"path": upload_result[0], "fileCount": upload_result[1], "operation": "append",
                    "tableDef": _get_structure(data_frame)}

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/adx/data-operation?tableName={table_name}"
    logger.info("Sending request: POST %s", url)
    logger.info("Sending request to ingest %s files from %s", str(upload_result[1]), upload_result[0])

    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status
    elif response.status_code == 200:
        _upsert_entities_affected_count(api_inputs=api_inputs, entities_affected_count=data_frame.shape[0])

    response_df = pandas.read_json(response.text, typ='series').to_frame().T
    response_df.columns = _column_name_cap(columns=response_df.columns)

    logger.info("API Response: %s", response_status)
    logger.info("Ingestion complete. ")

    return response_status, response_df


def upsert_file_row_count(api_inputs: ApiInputs, row_count: int):
    """Updates data feed file status row count.

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    row_count : number
        Number of rows

    Returns
    -------
    str
        Response status as a string.
    """

    if row_count is None:
        row_count = 0

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if (api_inputs.data_feed_id == '00000000-0000-0000-0000-000000000000' or
            api_inputs.data_feed_file_status_id == '00000000-0000-0000-0000-000000000000'):
        logger.error("upsert_file_row_count() can only be called in Production.")
        return False

    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
          f"{api_inputs.data_feed_id}/file-status/{api_inputs.data_feed_file_status_id}/row-count/{row_count}"

    response = requests.request("PUT", url, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status

    return response_status


def upsert_event_work_order_id(api_inputs: ApiInputs, event_task_id: uuid.UUID, integration_id: str,
                               work_order_status: str):
    """

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    event_task_id : uuid.UUID
        The value of the `work_order_input['EventTaskId']`
    integration_id : str
        The 3rd Party work order system's unique identifier for the work order
    work_order_status : str
        The status of the work order

    Returns
    -------
    (str, str)
        response_status, response.text - The response status of the call and the text from the response body.

    """

    if api_inputs.api_projects_endpoint == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return 'Error', 'You must call initialize() before using the API.'

    header = api_inputs.api_headers.default

    payload = {
        "IntegrationId": integration_id,
        "Status": work_order_status
    }

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/events/{str(event_task_id)}/work-order"

    response = requests.put(url=url, json=payload, headers=header)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, response.text

    return response_status, response.text


@_with_func_attrs(df_required_columns=['Identifier'])
def upsert_tags(api_inputs: ApiInputs, df: pandas.DataFrame, tag_level: TAG_LEVEL):
    """
    Upsert tags to Site/Device/Sensors as specified by the tag_level argument.
    
    Required fields are:
    - Identifier
    - Additional columns as TagGroups / Tags

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    df : DataFrame
        List of Devices along with corresponding TagsJson to upsert
    tag_level : TAG_LEVEL
        Level of tagging applied to the list of Identifier input.
            If tag_level is Site, Identifier should be InstallationIds.
            If tag_level is Device, Identifier should be DeviceIds.
            If tag_level is Sensor, Identifier should be ObjectPropertyIds.

    Returns
    -------
    List of affected records
    
    """
    data_frame = df.copy()

    # validate inputs
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    # validate df
    if data_frame.empty:
        logger.error("Data Frame is empty. Nothing to upsert.")
        return pandas.DataFrame()

    # validate tag level
    if not set([tag_level]).issubset(set(TAG_LEVEL.__args__)):
        logger.error('tag_level parameter must be set to one of the allowed values defined by the '
                        'TAG_LEVEL literal: %s', TAG_LEVEL.__args__)
        return pandas.DataFrame()

    required_columns = getattr(upsert_tags, 'df_required_columns')
    req_cols = ', '.join(required_columns)
    proposed_columns = data_frame.columns.tolist()

    # check if required columns are present
    if not set(required_columns).issubset(proposed_columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'integration.upsert_tags() - data_frame must contain the following columns: ' + req_cols

    slice_columns = set(proposed_columns).difference(set(required_columns))

    if len(list(slice_columns)) > 0:
        def update_values(row):
            j_row = row[list(slice_columns)].to_json()
            return str(j_row)

        data_frame['TagsJson'] = data_frame[list(slice_columns)].assign(
            **data_frame[list(slice_columns)].select_dtypes(['datetime', 'object']).astype(str)).apply(update_values, axis=1)
        data_frame = data_frame.drop(columns=list(slice_columns))
    else:
        logger.error("Additional Columns aside from Identifier is required to set as TagsJson.")
        return pandas.DataFrame()

    switcher_tag_level_identifier = {
        'Site': 'InstallationId',
        'Device': 'DeviceId',
        'Sensor': 'ObjectPropertyId'
    }

    tag_level_identifier_id = switcher_tag_level_identifier.get(tag_level)

    data_frame.rename(columns = {'Identifier':tag_level_identifier_id}, inplace = True)

    # action request
    headers = api_inputs.api_headers.integration

    switcher_tag_level_uri = {
        'Site': f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/installations/upsert-ingestion-tags",
        'Device': f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/devices/upsert-ingestion-tags",
        'Sensor': f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/sensors/upsert-ingestion-tags",
    }

    url = switcher_tag_level_uri.get(tag_level)

    logger.info("Sending request: POST %s", url)
    response = requests.post(url, data=data_frame.to_json(orient='records'), headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)

    logger.info("Response status: %s", response_status)

    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                        response.reason)
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)

    response_df = pandas.read_json(response.text)
    response_df.columns = _column_name_cap(response_df.columns)

    return response_df


@_with_func_attrs(df_required_columns=['DeviceId'])
def upsert_device_metadata(api_inputs: ApiInputs, df: pandas.DataFrame):
    """
    Upsert metadata on created devices.

    Required fields are:
    - DeviceId
    - Additional columns as Metadata

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    df : DataFrame
        List of Devices along with corresponding MetadataJson to upsert

    Returns
    -------
    List of affected Devices
    """
    data_frame = df.copy()

    # validate inputs
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    # validate df
    if data_frame.empty:
        logger.error("Data Frame is empty. Nothing to upsert.")
        return pandas.DataFrame()

    required_columns = getattr(upsert_device_metadata, 'df_required_columns')

    req_cols = ', '.join(required_columns)
    proposed_columns = data_frame.columns.tolist()

    # check if required columns are present
    if not set(required_columns).issubset(proposed_columns):
        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
        return 'integration.upsert_device_metadata() - data_frame must contain the following columns: ' + req_cols

    slice_columns = set(proposed_columns).difference(set(required_columns))

    if len(list(slice_columns)) > 0:
        def update_values(row):
            j_row = row[list(slice_columns)].to_json()
            return str(j_row)

        data_frame['MetadataJson'] = data_frame[list(slice_columns)].assign(
            **data_frame[list(slice_columns)].select_dtypes(['datetime', 'object']).astype(str)).apply(update_values, axis=1)
        data_frame = data_frame.drop(columns=list(slice_columns))
    else:
        logger.error("Additional Columns aside from DeviceId is required to set as MetadataJson.")
        return pandas.DataFrame()

    # action request
    headers = api_inputs.api_headers.integration

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/devices/upsert-ingestion-metadata"

    logger.info("Sending request: POST %s", url)
    response = requests.post(url, data=data_frame.to_json(orient='records'), headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)

    logger.info("Response status: %s", response_status)

    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                        response.reason)
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)

    response_df = pandas.read_json(response.text)
    response_df.columns = _column_name_cap(response_df.columns)

    return response_df



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for Cache data requests
"""

from datetime import datetime
import json
import logging
import sys
import requests
import uuid

import pandas

from .._utils._utils import (ApiInputs, is_valid_uuid)
from .._utils._constants import CACHE_SCOPE


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


def set_cache(api_inputs: ApiInputs, scope: CACHE_SCOPE, key: str, val: any, scope_id: uuid.UUID = None):
    """
    Sets data to be stored in cache for later retrieval.

    Parameters
    ----------
    api_inputs: ApiInputs
        Object returned by initialize() function.
    scope: CACHE_SCOPE
        Cache scope for the stored data
    key: str
        Key of to be stored data. Used when retrieving this data.
    val: any
        Data to be stored for later retrieval.
    scope_id: UUID
        UUID in relation to the scope that was set. Used as well when retrieving the data later on.
        For Task scope provide TaskId (self.id when calling from the driver)
        For DataFeed scope provide UUID4 for local testing. 
            api_inputs.data_feed_id will be used when running in the cloud. 
        For Portfolio scope, scope_id will be ignored and api_inputs.api_project_id will be used.

    Returns
    -------
    """
    
    if not set([scope]).issubset(set(CACHE_SCOPE.__args__)):
        logger.error('scope parameter must be set to one of the allowed values defined by the '
                        'CACHE_SCOPE literal: %s', CACHE_SCOPE.__args__)
        return pandas.DataFrame()
    
    scope_id_str = str(scope_id)
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if scope != 'Portfolio' and not is_valid_uuid(scope_id_str):
        logger.error("Invalid UUID v4 passed into scope_id parameter. Please provide a valid UUID v4.")
        return pandas.DataFrame()
    
    if isinstance(val, pandas.DataFrame):
        val = val.to_json(orient="records")
        memory_usage = sys.getsizeof(val)
        if (memory_usage > 2000000):
            logger.error(f"Data size limit reached. Only 2mb data size is allowed for cache. Size reached {round(memory_usage / 1000000, 2)}mb.")
            return pandas.DataFrame()
    elif isinstance(val, datetime):
        val = val.isoformat()

    headers = api_inputs.api_headers.integration
    
    json_payload = {
        "key": key.strip(),
        "value": val,
        "settings": {
            "dataFeedId": scope_id_str if api_inputs.data_feed_id is None \
                or api_inputs.data_feed_id == '00000000-0000-0000-0000-000000000000' \
                    or api_inputs.data_feed_id == '' else api_inputs.data_feed_id,
            "taskId": scope_id_str,
            "cacheScope": scope
        }
    }

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/task-insights/cache/set"
    logger.info("Sending request: POST %s", url)

    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    response_json = json.loads(response.text)

    logger.info("Response status: %s", response_status)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                        response_json['message'])
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)

    return response_json
    
    

def get_cache(api_inputs: ApiInputs, scope: CACHE_SCOPE, key: str, scope_id: uuid.UUID = None):
    """
    Gets data stored in cache.

    Parameters
    ----------
    api_inputs: ApiInputs
        Object returned by initialize() function.
    scope: CACHE_SCOPE
        Cache scope for the stored data
    key: str
        Key of to be stored data. Used when retrieving this data.
    scope_id: UUID
        UUID in relation to the scope that was set. Used as well when retrieving the data later on.
        For Task scope provide TaskId (self.id when calling from the driver)
        For DataFeed scope provide UUID4 for local testing. 
            api_inputs.data_feed_id will be used when running in the cloud. 
        For Portfolio scope, scope_id will be ignored and api_inputs.api_project_id will be used.

    Returns
    -------
    """

    if not set([scope]).issubset(set(CACHE_SCOPE.__args__)):
        logger.error('scope parameter must be set to one of the allowed values defined by the '
                        'CACHE_SCOPE literal: %s', CACHE_SCOPE.__args__)
        return pandas.DataFrame()

    scope_id_str = str(scope_id)
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if scope != 'Portfolio' and not is_valid_uuid(scope_id_str):
        logger.error("Invalid UUID v4 passed into scope_id parameter. Please provide a valid UUID v4.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.integration

    json_payload = {
        "key": key.strip(),
        "settings": {
            "dataFeedId": scope_id_str if api_inputs.data_feed_id is None \
                or api_inputs.data_feed_id == '00000000-0000-0000-0000-000000000000' \
                    or api_inputs.data_feed_id == '' else api_inputs.data_feed_id,
            "taskId": scope_id_str,
            "cacheScope": scope
        }
    }

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/task-insights/cache/get"
    logger.info("Sending request: POST %s", url)

    response = requests.post(url, json=json_payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    response_json = json.loads(response.text)
    
    logger.info("Response status: %s", response_status)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                        response_json['message'])
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
    
    return response_json
    



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for interacting with Data Sets from the Switch Automation Platform.

"""
import pandas
import requests
import logging
import sys
from io import StringIO
from .._utils._utils import ApiInputs
from .._utils._constants import DATA_SET_QUERY_PARAMETER_TYPES


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)



def get_folders(api_inputs: ApiInputs, folder_type="dataset"):
    """Get dataset folders

    Retrieves the list of DataSets folders for the given portfolio from the Switch Automation
    platform.

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    folder_type : str
        Type of folder records to retrieve. (Default value = "dataset")

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response, df) - Returns the response status of the call & a dataframe containing the data returned by the call.

    """

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    payload = {}
    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/folders?type={folder_type}"
    logger.info("Sending request: GET %s", url)

    response = requests.request("GET", url, data=payload, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    df = pandas.read_json(response.text)
    logger.info("API Call successful.")
    return response_status, df


def get_datasets_list(api_inputs: ApiInputs, include_path: bool = False, path_prefix="*"):
    """Get list of data sets.

    Retrieves the list of data set queries either for the entire project across all folders
    (if `include_path`=False) or specific to the folder designated in the `path_prefix`
    when `include_path` = True.

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    include_path : bool, default = False
        Include the folder path (Default value = False).
    path_prefix : str
        The folder path where queries should be returned from (Default value = "*").

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response, df) - Returns the response status of the call & a dataframe containing the data returned by the call.

    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    payload = {}
    headers = api_inputs.api_headers.default

    if path_prefix != "*":
        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/datasets?pathprefix=" \
              f"{path_prefix.replace('/', '|')}&includepath={str(include_path)}"
    else:
        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/datasets?includepath=" \
              f"{str(include_path)}"

    logger.info("Sending request: GET %s", url)

    response = requests.request("GET", url, data=payload, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    df = pandas.read_json(response.text)
    logger.info("API Call successful.")

    return response_status, df


def get_data(api_inputs: ApiInputs, dataset_id, parameters: list[dict] = None):
    """Retrieve data for the given Data Set.

    Retrieves data from the Switch Automation Platform for the given data set.

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    dataset_id : uuid.UUID
        The identifier for the data set.
    parameters : list[dict], Optional
        Any parameters required to be passed to the data set.
        Dict must contain `name`, `value`, and `type` items 

    Returns
    -------
    tuple[str, pandas.DataFrame]
        (response, df) - Returns the response status of the call & a dataframe containing the data returned by the call.
    """

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/datasets/queries/{str(dataset_id)}/data"
    logger.info("Sending request: POST %s", url)
    logger.info('DataSets: Sending request for DatasetID: %s', dataset_id)

    if parameters is None or len(parameters) <= 0:
        parameters = []
    else:
        for param in parameters:
            if not 'name' in param:
                logger.error(f"Missing 'name' property in parameter {param}.")
                return pandas.DataFrame()
            if not 'value' in param:
                logger.error(f"Missing 'value' property in parameter {param}.")
                return pandas.DataFrame()
            if not 'type' in param:
                logger.error(f"Missing 'type' property in parameter {param}.")
                return pandas.DataFrame()
            if not param['type'] in DATA_SET_QUERY_PARAMETER_TYPES.__args__:
                logger.error('type parameter must be set to one of the allowed values defined by the '
                        'DATA_SET_QUERY_PARAMETER_TYPES literal: %s', DATA_SET_QUERY_PARAMETER_TYPES.__args__)
                return pandas.DataFrame()

    payload = {
        'parameters': parameters
    }

    response = requests.post(url, json=payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('DataSets: no data returned for this call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    df = pandas.read_json(StringIO(response.text), dtype=False)
    logger.info("API Call successful.")

    return response_status, df



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module containing the helper functions useful when integrating asset creation, asset updates, data ingestion, etc
into the Switch Automation Platform.
"""
import json
import pandas
import requests
import datetime
import logging
import sys
from .._utils._constants import SUPPORT_PAYLOAD_TYPE  # , QUERY_LANGUAGE
from .._utils._utils import ApiInputs, requests_retry_session, is_valid_uuid

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


def _timezones(api_inputs: ApiInputs):
    """Get timezones

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.

    Returns
    -------
    df : pandas.DataFrame


    """
    # payload = {}
    headers = api_inputs.api_headers.default

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    # upload Blobs to folder
    url = f"{api_inputs.api_base_url}/timezones/all"
    logger.info("Sending request: GET %s", url)

    response = requests.request("GET", url, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    df = pandas.read_json(response.text, orient='index')
    df.rename(columns={0: 'TimezoneName'}, inplace=True)
    df['TimezoneId'] = df.index

    return df


def _timezone_offsets(api_inputs: ApiInputs, date_from: datetime.date, date_to: datetime.date,
                      installation_id_list: list):
    """Get timezones

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    date_from : datetime.date
        First (earliest) record
    date_to : datetime.date
        Last (latest) record
    installation_id_list : list
        List of InstallationIDs

    Returns
    -------
    df : pandas.DataFrame
        A dataframe containing the timezone offsets per InstallationID - if multiple timezone offsets occur for the
        submitted period, there will be a row per offset date range. E.g. for a timezone with daylight savings

    """
    payload = {
        "dateFrom": date_from.isoformat(),
        "dateTo": date_to.isoformat(),
        "installationIds": installation_id_list
    }

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_base_url}/timezones/offsets"
    logger.info("Sending request: POST %s", url)

    try:
        response = requests_retry_session(method_whitelist=['POST']).post(url, json=payload, headers=headers)

        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        response_data_frame = pandas.read_json(response.text)
        return response_data_frame
    except Exception as ex:
        logger.error("API Call was not successful.", ex)
        return pandas.DataFrame()


def _timezone_dst_offsets(api_inputs: ApiInputs, date_from: datetime.date, date_to: datetime.date,
                          installation_id_list: list = None, timezone_name: str = None):
    """Get timezones

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    date_from : datetime.date
        First (earliest) record
    date_to : datetime.date
        Last (latest) record
    installation_id_list : list
        List of InstallationIDs
    timezone_name: str
        Specific timezone name to retrieve DST offsets.
        Defaults to None.

    Returns
    -------
    df : pandas.DataFrame
        A dataframe containing the timezone offsets per InstallationID - if multiple timezone offsets occur for the
        submitted period, there will be a row per offset date range. E.g. for a timezone with daylight savings

    """
    payload = {
        "dateFrom": date_from.isoformat(),
        "dateTo": date_to.isoformat(),
        "installationIds": installation_id_list,
        "timezoneName": timezone_name
    }

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if timezone_name == '':
        logger.error("timezone_name parameter cannot be an empty string.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_base_url}/timezones/dst-transitions"
    if not timezone_name is None:
        url = f"{api_inputs.api_base_url}/timezones/name/dst-transitions"

    logger.info("Sending request: POST %s", url)

    try:
        response = requests_retry_session(method_whitelist=['POST']).post(url, json=payload, headers=headers)

        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        timezone_intervals = json.loads(response.text)

        if not set(['installationTimezoneList','dstTransitions', 'dstIntervals']).issubset(timezone_intervals.columns):
            logger.error('No Timezone DST Intervals returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        site_timezones = pandas.DataFrame.from_dict(timezone_intervals['installationTimezoneList'])
        timezone_offsets = pandas.DataFrame.from_dict(timezone_intervals['dstTransitions'])
        timezone_offsets = timezone_offsets.explode('dstIntervals')
        timezone_offsets = timezone_offsets.reset_index(drop=True).join(timezone_offsets.reset_index(drop=True)
                                                                        .dstIntervals.apply(pandas.Series)).drop(columns=['dstIntervals', 'timezoneName'])
        if timezone_name is None:
            return site_timezones.merge(timezone_offsets, on='timezoneId')
        else:
            return timezone_offsets
    except Exception as ex:
        logger.error("API Call was not successful.", ex)
        return pandas.DataFrame()

def _upsert_entities_affected_count(api_inputs: ApiInputs, entities_affected_count: int):
    """Updates data feed and data feed file status entities affected count.

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    entities_affected_count : int
        Count of affected entities - i.e. how many records are being upserted, replaced or appended.

    Returns
    -------

    """

    if entities_affected_count is None:
        entities_affected_count = 0

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if not is_valid_uuid(api_inputs.data_feed_id):
        logger.error("Entities Affected Count can only be upserted when called in Production.")
        return False

    if not is_valid_uuid(api_inputs.data_feed_file_status_id):
        logger.error("Entities Affected Count can only be upserted when called in Production.")
        return False

    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
          f"{api_inputs.data_feed_id}/file-status/{api_inputs.data_feed_file_status_id}/entities-affected/" \
          f"{entities_affected_count}"

    response = requests.request("PUT", url, timeout=20, headers=headers)

    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
    elif response.status_code == 200:
        logger.info("Entities Affected Count successfully upserted. ")


def _adx_support(api_inputs: ApiInputs, payload_type: SUPPORT_PAYLOAD_TYPE):
    """
        Call ADX Support endpoint to trigger SQL->ADX sync

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    payload_type : SUPPORT_PAYLOAD_TYPE
        Payload Type to sync ADX

    Returns
    -------


    """
    headers = api_inputs.api_headers.default

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if payload_type == None:
        logger.error("You must provide the ADX Support Payload Type to SYNC ADX.")
        return pandas.DataFrame()

    if not set([payload_type]).issubset(set(SUPPORT_PAYLOAD_TYPE.__args__)):
        logger.error('payload_type parameter must be set to one of the allowed values defined by the '
                     'SUPPORT_PAYLOAD_TYPE literal: %s', SUPPORT_PAYLOAD_TYPE.__args__)
        return pandas.DataFrame()

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/adx/sync/{payload_type}"

    response = requests.request("PUT", url, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    logger.info("API Call successful.")

    return response_status, response.text



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module defining the various definitions used internally by the pipeline module classes.
"""
# import uuid
from typing import List
from .._utils._utils import convert_to_pascal_case
from .._utils._constants import INTEGRATION_SETTINGS_EDITORS
# from ..integration import get_templates, get_units_of_measure
# from .. import initialize


class BaseProperty:
    """BaseProperty class definition

    Parameters
    ----------
    property_name : str
        Must be in pascal case
    display_label : str
        Pretty version of the property_name
    editor : INTEGRATION_SETTINGS_EDITORS
        The editor used in the UI
    default_value : , default=None
        The default value, if applicable, for the given property
    allowed_values : list, default=None
        the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be None.

    """
    def __init__(self, property_name: str, display_label: str, editor: INTEGRATION_SETTINGS_EDITORS,
                 default_value: object = None, allowed_values: List[object] = None):
        """
        Parameters
        ----------
        property_name : str
            Must be in pascal case
        display_label : str
            Pretty version of the property_name
        editor : INTEGRATION_SETTINGS_EDITORS
            The editor used in the UI
        default_value : , default=None
            The default value, if applicable, for the given property
        allowed_values : list, default=None
            the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be
            None.

        """
        self.property_name = property_name
        self.display_label = display_label
        self.editor = editor
        self.default_value = default_value
        self.allowed_values = allowed_values

    @property
    def property_name(self) -> str:
        """property_name : str
            Must be in pascal case"""
        return self._property_name

    @property_name.setter
    def property_name(self, value) -> str:
        """property_name : str
                    Must be in pascal case"""
        if type(value) != str:
            raise TypeError(f"The property_name parameter only accepts strings. ")
        elif type(value) == str and not 0 < len(value) < 50:
            raise ValueError(
                f"property_name length must be between 1 and 50. Supplied '{value}' "
                f"contains {len(value)} characters. ")
        elif type(value) == str and 0 < len(value) < 50 and value != convert_to_pascal_case(value):
            raise ValueError(
                f"property_name must be in pascal case - i.e. start with a capital letter and contain no spaces, "
                f"dashes, underscores, etc. The supplied '{value}' is not in pascal case. "
            )
        self._property_name = value

    @property
    def display_label(self) -> str:
        """The display label for the given property"""
        return self._display_label

    @display_label.setter
    def display_label(self, value) -> str:
        """The display label for the given property"""
        if type(value) != str:
            raise TypeError(f"The display_label parameter only accepts strings. ")
        elif type(value) == str and not 0 < len(value) <= 255:
            raise ValueError(
                f"display_label length must be between 1 and 50. Supplied '{value}' "
                f"contains {len(value)} characters. ")
        self._display_label = value

    @property
    def editor(self) -> INTEGRATION_SETTINGS_EDITORS:
        """The editor to be used in the UI"""
        return self._editor

    @editor.setter
    def editor(self, value) -> INTEGRATION_SETTINGS_EDITORS:
        """The editor to be used in the UI"""
        if not set([value]).issubset(set(INTEGRATION_SETTINGS_EDITORS.__args__)):
            raise ValueError(
                f"Supplied value '{value}' is invalid. The editor parameter must be set to one of the allowed values "
                f"defined by the INTEGRATION_SETTINGS_EDITORS literal: {INTEGRATION_SETTINGS_EDITORS.__args__}")
        self._editor = value


class EventWorkOrderFieldDefinition(BaseProperty):
    """EventWorkOrderFieldDefinition class definition

    Used to define the set of fields available in Events UI for creation of work orders in 3rd party systems.

    Parameters
    ----------
    property_name : str
        Must be in pascal case
    display_label : str
        Pretty version of the property_name (what is displayed in Events UI).
    editor : INTEGRATION_SETTINGS_EDITORS
        The editor used in the UI
    default_value : , default=None
        The default value, if applicable, for the given property
    allowed_values : list, default=None
        the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be None.
    """
    def __init__(self, property_name, display_label, editor, default_value, allowed_values):
        """

        Parameters
        ----------
        property_name : str
            Must be in pascal case
        display_label : str
            Pretty version of the property_name
        editor : INTEGRATION_SETTINGS_EDITORS
            The editor used in the UI
        default_value : , default=None
            The default value, if applicable, for the given property
        allowed_values : list, default=None
            The set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be
            None.
        """
        super(EventWorkOrderFieldDefinition, self).__init__(
            property_name=property_name, display_label=display_label, editor=editor, default_value=default_value,
            allowed_values=allowed_values)


class AnalyticsSettings(BaseProperty):
    """

    """
    def __init__(self, property_name, display_label, editor, default_value, allowed_values):
        """
        Parameters
        ----------
        property_name : str
            Must be in pascal case
        display_label : str
            Pretty version of the property_name
        editor : INTEGRATION_SETTINGS_EDITORS
            The editor used in the UI
        default_value : , default=None
            The default value, if applicable, for the given property
        allowed_values : list, default=None
            the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be
            None.
        """
        super(AnalyticsSettings, self).__init__(
            property_name=property_name, display_label=display_label, editor=editor, default_value=default_value,
            allowed_values=allowed_values)


class IntegrationSettings(BaseProperty):
    """Class used for defining the UI editors, default values, etc for the key-values passed to integration_settings
    dictionaries.

    Parameters
    ----------
    property_name : str
        Must be in pascal case
    display_label : str
        Pretty version of the property_name
    editor : INTEGRATION_SETTINGS_EDITORS
        The editor used in the UI
    default_value : , default=None
        The default value, if applicable, for the given property
    allowed_values : list, default=None
        the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be
        None.
    """
    def __init__(self, property_name, display_label, editor, default_value, allowed_values):
        """

        Parameters
        ----------
        property_name : str
            Must be in pascal case
        display_label : str
            Pretty version of the property_name
        editor : INTEGRATION_SETTINGS_EDITORS
            The editor used in the UI
        default_value : , default=None
            The default value, if applicable, for the given property
        allowed_values : list, default=None
            the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be
            None.
        """
        super(IntegrationSettings, self).__init__(
            property_name=property_name, display_label=display_label, editor=editor, default_value=default_value,
            allowed_values=allowed_values)


# class IntegrationDevicePropertyDefinition2:
#     def __init__(self, property_name: str, display_label: str, default_value: object = None,
#                  editor: INTEGRATION_SETTINGS_EDITORS = 'text_box', values: List[object] = None,
#                  required_for_task: bool = False, use_for_discovery: bool = False, editable_in_discovery: bool = False):
#         self.property_name = property_name
#         self.display_label = display_label
#         self.default_value = default_value
#         self.editor = editor
#         self.values = values
#         self.required_for_task = required_for_task
#         self.use_for_discovery = use_for_discovery
#         self.editable_in_discovery = editable_in_discovery
#
#     @property
#     def property_name(self):
#         return self._property_name
#
#     @property_name.setter
#     def property_name(self, value):
#         if type(value) != str:
#             raise TypeError(f"The property_name parameter only accepts strings. ")
#         elif type(value) == str and not 0 < len(value) < 50:
#             raise ValueError(
#                 f"property_name length must be between 1 and 50. Supplied '{value}' "
#                 f"contains {len(value)} characters. ")
#         elif type(value) == str and 0 < len(value) < 50 and value != convert_to_pascal_case(value):
#             raise ValueError(
#                 f"property_name must be in pascal case - i.e. start with a capital letter and contain no spaces, "
#                 f"dashes, underscores, etc. The supplied '{value}' is not in pascal case. "
#             )
#         self._property_name = value
#
#     @property
#     def display_label(self):
#         return self._display_label
#
#     @display_label.setter
#     def display_label(self, value):
#         if type(value) != str:
#             raise TypeError(f"The display_label parameter only accepts strings. ")
#         elif type(value) == str and not 0 < len(value) <= 255:
#             raise ValueError(
#                 f"display_label length must be between 1 and 50. Supplied '{value}' "
#                 f"contains {len(value)} characters. ")
#         self._display_label = value
#
#     @property
#     def editor(self):
#         return self._editor
#
#     @editor.setter
#     def editor(self, value):
#         if not set([value]).issubset(set(INTEGRATION_SETTINGS_EDITORS.__args__)):
#             raise ValueError(
#                 f"The editor parameter must be set to one of the allowed values defined by the "
#                 f"INTEGRATION_SETTINGS_EDITORS literal: {INTEGRATION_SETTINGS_EDITORS.__args__}")
#         self._editor = value
#
#     @property
#     def required_for_task(self):
#         return self._required_for_task
#
#     @required_for_task.setter
#     def required_for_task(self, value):
#         if type(value) != bool:
#             raise TypeError(f"The required_for_task parameter must be passed a boolean. ")
#         self._required_for_task = value
#
#     @property
#     def use_for_discovery(self):
#         return self._use_for_discovery
#
#     @use_for_discovery.setter
#     def use_for_discovery(self, value):
#         if type(value) != bool:
#             raise TypeError(f"The use_for_discovery parameter must be passed a boolean. ")
#         self._use_for_discovery = value
#
#     @property
#     def editable_in_discovery(self):
#         return self._editable_in_discovery
#
#     @editable_in_discovery.setter
#     def editable_in_discovery(self, value):
#         if type(value) != bool:
#             raise TypeError(f"The editable_in_discovery parameter must be passed a boolean. ")
#         self._editable_in_discovery = value


class IntegrationDeviceConfigPropertyDefinition(BaseProperty):
    """
    Class corresponding to the configuration properties for the Integration device type for the given task.

    Parameters
    ----------
    property_name : str
        Must be in pascal case
    display_label : str
        Pretty version of the property_name
    editor : INTEGRATION_SETTINGS_EDITORS
        The editor used in the UI
    default_value : , default=None
        The default value, if applicable, for the given property
    allowed_values : list, default=None
        the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be
        None.
    required_for_task : bool, default=False
        Determines whether the given property is required for the task to run (Default Value = False).
    use_for_discovery: bool, default=False
        Determines whether the given property is required for the discovery to run (Default Value = False).
    editable_in_discovery: bool, default=False
        Determines whether the given property is editable when triggering discovery (Default value = False).
    """
    def __init__(self, property_name, display_label, editor, default_value, allowed_values,
                 required_for_task: bool = False, use_for_discovery: bool = False, editable_in_discovery: bool = False):
        """

        Parameters
        ----------
        property_name : str
            Must be in pascal case
        display_label : str
            Pretty version of the property_name
        editor : INTEGRATION_SETTINGS_EDITORS
            The editor used in the UI
        default_value : , default=None
            The default value, if applicable, for the given property
        allowed_values : list, default=None
            the set of allowed values (if applicable) for the given property_name. If editor=text_box, this should be
            None.
        required_for_task : bool, default=False
            Determines whether the given property is required for the task to run (Default Value = False).
        use_for_discovery: bool, default=False
            Determines whether the given property is required for the discovery to run (Default Value = False).
        editable_in_discovery: bool, default=False
            Determines whether the given property is editable when triggering discovery (Default value = False).
        """
        super(IntegrationDeviceConfigPropertyDefinition, self).__init__(
            property_name=property_name, display_label=display_label, editor=editor, default_value=default_value,
            allowed_values=allowed_values)
        self.required_for_task = required_for_task
        self.use_for_discovery = use_for_discovery
        self.editable_in_discovery = editable_in_discovery
        self._is_configuration = True

    @property
    def required_for_task(self) -> bool:
        """"Determines whether the given property is required for the task to run (Default Value = False). """
        return self._required_for_task

    @required_for_task.setter
    def required_for_task(self, value) -> bool:
        """Determines whether the given property is required for the task to run. """
        if type(value) != bool:
            raise TypeError(f"The required_for_task parameter must be passed a boolean. ")
        self._required_for_task = value

    @property
    def use_for_discovery(self) -> bool:
        """Defines whether the given property should be used when triggering a discovery. """
        return self._use_for_discovery

    @use_for_discovery.setter
    def use_for_discovery(self, value: bool) -> bool:
        """Defines whether the given property should be used when triggering a discovery. """
        if type(value) != bool:
            raise TypeError(f"The use_for_discovery parameter must be passed a boolean. ")
        self._use_for_discovery = value

    @property
    def editable_in_discovery(self) -> bool:
        """Defines whether the given property is editable when triggering a discovery. """
        return self._editable_in_discovery

    @editable_in_discovery.setter
    def editable_in_discovery(self, value) -> bool:
        """Defines whether the given property is editable when triggering a discovery. """
        if type(value) != bool:
            raise TypeError(f"The editable_in_discovery parameter must be passed a boolean. ")
        self._editable_in_discovery = value

    @property
    def is_configuration(self):
        return self._is_configuration


class IntegrationDeviceDefinition:
    """
    Aligns with the definition of the Integration DriverDeviceType.
    """
    def __init__(self, device_type: str, default_device_name: str,
                 config_properties: List[IntegrationDeviceConfigPropertyDefinition],
                 expose_address: bool, address_label: str):
        """

        Parameters
        ----------
        device_type : str
            The DriverDeviceType of the integration device type. Must be in pascal-case.
        default_device_name : str
            The default device name for this DriverDeviceType.
        config_properties : List[IntegrationDeviceConfigPropertyDefinition]
            A list of the configuration properties associated with the device type. The items in the list should be
            instances of the `IntegrationDeviceConfigPropertyDefinition` class.
        expose_address : bool
            Defines whether the Address is exposed to the user & platform UI.
        address_label : str or None
            If `expose_address` is True, tben this defines the label shown to user in UI.
        """
        self.device_type = device_type
        self.default_device_name = default_device_name
        self.config_properties = config_properties
        self._is_integration = True
        self.expose_address = expose_address
        self.address_label = address_label

    @property
    def device_type(self) -> str:
        """The DriverDeviceType of the integration device type. Must be in pascal-case. """
        return self._device_type

    @device_type.setter
    def device_type(self, value) -> str:
        """The DriverDeviceType of the integration device type. Must be in pascal-case. """
        if type(value) != str:
            raise TypeError(f"The device_type parameter only accepts strings. ")
        elif type(value) == str and not 0 < len(value) <= 100:
            raise ValueError(f"DeviceTypeDefinition.device_type length must be between 1 and 100. Supplied '{value}' "
                             f"contains {len(value)} characters. ")
        self._device_type = value

    @property
    def default_device_name(self) -> str:
        """The default device name for this DriverDeviceType. """
        return self._default_device_name

    @default_device_name.setter
    def default_device_name(self, value) -> str:
        """The default device name for this DriverDeviceType. """
        if type(value) != str:
            raise TypeError(f"The default_device_name parameter only accepts strings. ")
        elif type(value) == str and not 0 < len(value) < 50:
            raise ValueError(f"IntegrationDeviceTypeDefinition.default_device_name length must be between 1 and 50. "
                             f"Supplied '{value}' contains {len(value)} characters. ")
        self._default_device_name = value

    @property
    def config_properties(self) -> List[IntegrationDeviceConfigPropertyDefinition]:
        """A list of the configuration properties associated with the device type. The items in the list should be
            instances of the `IntegrationDeviceConfigPropertyDefinition` class. """
        return self._config_properties

    @config_properties.setter
    def config_properties(self, value) -> List[IntegrationDeviceConfigPropertyDefinition]:
        """A list of the configuration properties associated with the device type. The items in the list should be
            instances of the `IntegrationDeviceConfigPropertyDefinition` class. """
        if type(value) != list:
            raise TypeError(f"The sensor_properties parameter must be passed a list. ")
        elif type(value) == list:
            for i in range(len(value)):
                if isinstance(value[i], IntegrationDeviceConfigPropertyDefinition) == False:
                    raise TypeError(f"Each item in the list passed to the config_properties parameter must be an "
                                    f"instance of the IntegrationDeviceConfigPropertyDefinition class")
        self._config_properties = value

    @property
    def is_integration(self):
        return self._is_integration

    @property
    def expose_address(self) -> bool:
        """Defines whether the Address is exposed to the user & platform UI. """
        return self._expose_address

    @expose_address.setter
    def expose_address(self, value) -> bool:
        """Defines whether the Address is exposed to the user & platform UI. """
        if type(value) != bool:
            raise TypeError(f"The expose_address parameter must be passed a boolean. ")
        self._expose_address = value

    @property
    def address_label(self) -> str:
        """If `expose_address` is True, tben this defines the label shown to user in UI."""
        return self._address_label

    @address_label.setter
    def address_label(self, value) -> str:
        """If `expose_address` is True, tben this defines the label shown to user in UI."""
        if type(value) != str and self.expose_address == False and value is not None:
            raise TypeError(f"The address_label parameter only accepts strings. ")
        elif value is None and self.expose_address == True:
            raise ValueError(f"address_label must be set to a string (in PascalCase) when expose_address is True. The "
                             f"parameter address_label can only be set to None when expose_address is False. ")
        elif type(value) == str and value != convert_to_pascal_case(value):
            raise ValueError(
                f"address_label must be in PascalCase - i.e. start with a capital letter and contain no spaces, "
                f"dashes, underscores, etc. The supplied '{value}' is not in PascalCase. "
            )
        self._address_label = value


# class SensorDefinition:
#     def __init__(self, property_name: str, default_sensor_name: str, object_property_template_name, unit_of_measure,
#                  is_configuration: bool = False):
#         self.property_name = property_name
#         # self.is_configuration = is_configuration
#         self._is_configuration = False
#         self.default_sensor_name = default_sensor_name
#         self.object_property_template_name = object_property_template_name
#         self.unit_of_measure = unit_of_measure
#
#     @property
#     def property_name(self):
#         return self._property_name
#
#     @property_name.setter
#     def property_name(self, value):
#         if type(value) != str:
#             raise TypeError(f"The property_name parameter only accepts strings. ")
#         elif type(value) == str and not 0 < len(value) < 50:
#             raise ValueError(
#                 f"property_name length must be between 1 and 50. Supplied '{value}' "
#                 f"contains {len(value)} characters. ")
#         elif type(value) == str and 0 < len(value) < 50 and value != convert_to_pascal_case(value):
#             raise ValueError(
#                 f"property_name must be in pascal case - i.e. start with a capital letter and contain no spaces, "
#                 f"dashes, underscores, etc. The supplied '{value}' is not in pascal case. "
#             )
#         self._property_name = value
#
#     @property
#     def is_configuration(self):
#         return self._is_configuration
#
#     # @is_configuration.setter
#     # def is_configuration(self, value):
#     #     if type(value) != bool:
#     #         raise TypeError(f"The is_configuration parameter only accepts boolean values - i.e. True or False ")
#     #     self._is_configuration = value
#
#     @property
#     def object_property_template_name(self):
#         return self._object_property_template_name
#
#     @object_property_template_name.setter
#     def object_property_template_name(self, value):  # max characters = 255, pascal case
#         allowed_values = get_templates(api_inputs=api_inputs)
#         if self.is_configuration == False and not set([value]).issubset(set(allowed_values[
#                                                                                 'ObjectPropertyTemplateName'])):
#             raise ValueError(f"{value} is not an allowed value for ObjectPropertyTemplateName. Please use "
#                              f"sw.integration.get_templates() function to retrieve list of the allowed values. ")
#         elif self.is_configuration == True and value is not None:
#             raise ValueError(f"If the defined property has IsConfiguration set to True, the value for "
#                              f"ObjectPropertyTemplateName should be set to None. ")
#         elif self.is_configuration == True and value is None:
#             self._object_property_type = None
#         elif self.is_configuration == False and set([value]).issubset(set(
#                 allowed_values['ObjectPropertyTemplateName'])):
#             self._object_property_type = allowed_values[
#                 allowed_values['ObjectPropertyTemplateName'] == value].ObjectPropertyType.item()
#         self._object_property_template_name = value
#
#     # @object_property_template_name.setter
#     # def object_property_template_name(self, value):
#     #     allowed_values = get_templates(api_inputs=api_inputs)
#     #     if self.is_configuration == False and not set([value]).issubset(set(allowed_values[
#     #                                                                             'ObjectPropertyTemplateName'])):
#     #         raise ValueError(f"{value} is not an allowed value for ObjectPropertyTemplateName. Please use "
#     #                          f"sw.integration.get_templates() function to retrieve list of the allowed values. ")
#     #     elif self.is_configuration==True and value is not None:
#     #         raise ValueError(f"If the defined property has IsConfiguration set to True, the value for "
#     #                          f"ObjectPropertyTemplateName should be set to None. ")
#     #     elif self.is_configuration==True and value is None:
#     #         self._object_property_type = None
#     #     elif self.is_configuration == False and set([value]).issubset(set(
#     #             allowed_values['ObjectPropertyTemplateName'])):
#     #         self._object_property_type = allowed_values[
#     #             allowed_values['ObjectPropertyTemplateName'] == value].ObjectPropertyType.item()
#     #     self._object_property_template_name = value
#
#     @property
#     def unit_of_measure(self):
#         return self._unit_of_measure
#
#     @unit_of_measure.setter
#     def unit_of_measure(self, value):  # max chars = 50, sentence case
#         if self._object_property_type is not None:
#             allowed_values = get_units_of_measure(api_inputs=initialize(),
#                                                   object_property_type=self._object_property_type)
#             if self.is_configuration == False and not set([value]).issubset(set(allowed_values[
#                                                                                     'UnitOfMeasureDescription'])):
#                 raise ValueError(f"{value} is not an allowed value for the Unit of Measure. Please use "
#                                  f"sw.integration.get_units_of_measure() function to retrieve list of the "
#                                  f"allowed values. ")
#             elif self.is_configuration == True and value is not None:
#                 raise ValueError(f"If the defined property has IsConfiguration set to True, the value for "
#                                  f"UnitOfMeasure should be set to None. ")
#         self._unit_of_measure = value
#       # if self._object_property_type is not None:
#       #       allowed_values = sw.integration.get_units_of_measure(api_inputs=api_inputs,
#       #                                                            object_property_type=self._object_property_type)
#       #       if self.is_configuration == False and not set([value]).issubset(set(allowed_values[
#       #                                                                               'UnitOfMeasureDescription'])):
#       #           raise ValueError(f"{value} is not an allowed value for the Unit of Measure. Please use "
#       #                            f"sw.integration.get_units_of_measure() function to retrieve list of the "
#       #                            f"allowed values. ")
#       #       elif self.is_configuration==True:
#       #           raise ValueError(f"If the defined property has IsConfiguration set to True, the value for "
#       #                            f"UnitOfMeasure should be set to None. ")
#       #   elif self._object_property_type is None:
#       #       if self.is_configuration == False:
#       #           raise ValueError(f"unit_of_measure can only be set to None if the is_configuration parameter is set"
#       #                            f" to True. ")
#
#     @property
#     def default_sensor_name(self):
#         return self._default_sensor_name
#
#     @default_sensor_name.setter
#     def default_sensor_name(self, value):
#         if type(value) != str:
#             raise TypeError(f"The default_sensor_name parameter only accepts strings. ")
#         elif type(value) == str and not 0 < len(value) <= 255:
#             raise ValueError(
#                 f"default_sensor_name length must be between 1 and 255. Supplied '{value}' "
#                 f"contains {len(value)} characters. ")
#         self._default_sensor_name = value
#
#
# class DeviceTypeConfigPropertyDefinition(BaseProperty):
#     def __init__(self, property_name, display_label, default_value,
#                  editor, allowed_values):
#         super(DeviceTypeConfigPropertyDefinition, self).__init__(property_name, display_label, default_value,
#                                                                  editor, allowed_values)
#         self._is_configuration = True
#         self._object_property_template_name = 'Configuration'
#         self._unit_of_measure = None
#
#     @property
#     def is_configuration(self):
#         return self._is_configuration
#
#     @property
#     def object_property_template_name(self):
#         return self._object_property_template_name
#
#     @property
#     def unit_of_measure(self):
#         return self._unit_of_measure
#
#
# class DeviceTypeDefinition:
#     def __init__(self, device_type: str, default_device_name: str, sensor_properties: List[SensorDefinition],
#                  expose_address: bool, address_label: str,
#                  config_properties: List[DeviceTypeConfigPropertyDefinition] = None):
#         self.device_type = device_type
#         self.default_device_name = default_device_name
#         self.sensor_properties = sensor_properties
#         self._is_integration = False
#         self.expose_address = expose_address
#         self.address_label = address_label
#         self.config_properties = config_properties
#
#     @property
#     def device_type(self):
#         return self._device_type
#
#     @device_type.setter
#     def device_type(self, value):
#         if type(value) != str:
#             raise TypeError(f"The device_type parameter only accepts strings. ")
#         elif type(value) == str and not 0 < len(value) < 50:
#             raise ValueError(f"DeviceTypeDefinition.device_type length must be between 1 and 50. Supplied '{value}' "
#                              f"contains {len(value)} characters. ")
#         elif type(value) == str and 0 < len(value) < 50 and value != convert_to_pascal_case(value):
#             raise ValueError(
#                 f"device_type must be in pascal case - i.e. start with a capital letter and contain no spaces, "
#                 f"dashes, underscores, etc. The supplied '{value}' is not in pascal case. "
#             )
#         self._device_type = value
#
#     @property
#     def default_device_name(self):
#         return self._default_device_name
#
#     @default_device_name.setter
#     def default_device_name(self, value):
#         if type(value) != str:
#             raise TypeError(f"The default_device_name parameter only accepts strings. ")
#         elif type(value) == str and len(value) > 50:
#             raise ValueError(f"The default_device_name has a maximum character count of 100. Supplied '{value}' "
#                              f"contains {len(value)} characters. ")
#         elif type(value) == str and len(value) == 0:
#             raise ValueError(f"The default_device_name must not be an empty string. ")
#         self._default_device_name = value
#
#     @property
#     def sensor_properties(self):
#         return self._sensor_properties
#
#     @sensor_properties.setter
#     def sensor_properties(self, value):
#         if type(value) != list:
#             raise TypeError(f"The sensor_properties parameter must be passed a list. ")
#         elif type(value) == list:
#             for i in range(len(value)):
#                 if isinstance(value[i], SensorDefinition) == False:
#                     raise TypeError(f"Each item in the list passed to the sensor_properties parameter must be an "
#                                     f"instance of the SensorDefinition class")
#         self._sensor_properties = value
#
#     @property
#     def is_integration(self):
#         return self._is_integration
#
#     @property
#     def expose_address(self):
#         return self._expose_address
#
#     @expose_address.setter
#     def expose_address(self, value):
#         if type(value) != bool:
#             raise TypeError(f"The expose_address parameter must be passed a boolean. ")
#         self._expose_address = value
#
#     @property
#     def address_label(self):
#         return self._address_label
#
#     @address_label.setter
#     def address_label(self, value):
#         if type(value) != str:
#             raise TypeError(f"The address_label parameter only accepts strings. ")
#         elif type(value) == str and value != convert_to_pascal_case(value):
#             raise ValueError(
#                 f"address_label must be in pascal case - i.e. start with a capital letter and contain no spaces, "
#                 f"dashes, underscores, etc. The supplied '{value}' is not in pascal case. "
#             )
#         self._address_label = value
#
#     @property
#     def config_properties(self):
#         return self._config_properties
#
#     @config_properties.setter
#     def config_properties(self, value):
#         if type(value) != list and value is not None:
#             raise TypeError(f"The config_properties parameter must be passed a list. ")
#         elif type(value) == list:
#             for i in range(len(value)):
#                 if isinstance(value[i], DeviceTypeConfigPropertyDefinition) == False:
#                     raise TypeError(f"Each item in the list passed to the config_properties parameter must be an "
#                                     f"instance of the DeviceTypeConfigPropertyDefinition class")
#         self._config_properties = value
#



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for integrating asset creation, asset updates, data ingestion, etc into the Switch Automation Platform.
"""

from .integration import (upsert_data, upsert_sites, upsert_workorders, upsert_device_sensors,
                          upsert_timeseries_ds, replace_data, append_data, upsert_file_row_count,
                          upsert_discovered_records, upsert_timeseries, upsert_tags, upsert_device_metadata)

from .helpers import (get_sites, get_tag_groups, get_metadata_keys, get_data, get_device_sensors, get_templates,
                      get_units_of_measure, get_states_by_country, get_operation_state, get_equipment_classes,
                      load_data, get_timezones)

__all__ = ['upsert_data', 'upsert_sites', 'upsert_workorders', 'upsert_device_sensors', 'upsert_timeseries_ds',
           'replace_data', 'append_data', 'get_sites', 'get_tag_groups', 'get_metadata_keys', 'get_data',
           'get_device_sensors', 'get_templates', 'get_units_of_measure', 'get_states_by_country',
           'get_operation_state', 'get_equipment_classes', 'load_data', 'upsert_file_row_count',
           'upsert_discovered_records', 'upsert_timeseries', 'upsert_tags', 'upsert_device_metadata', 'get_timezones']



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Miscellaneous constants, utility functions & interactions with storage accounts.
"""

# import ._utils
# import ._constants
# import ._platform



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module containing strategies for storing and retrieving credentials.
"""



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for .....
"""
# import datetime
import re
import uuid
from collections import namedtuple
from typing import Union, List
import string
import secrets
import pandas
import pandera
import logging
import sys
import requests
from requests.adapters import HTTPAdapter

from requests.sessions import session
from urllib3.util.retry import Retry
from .._utils._constants import WORK_ORDER_STATUS, WORK_ORDER_CATEGORY, WORK_ORDER_PRIORITY

__all__ = ['generate_password', 'convert_to_sqm', 'convert_to_pascal_case', 'ValidatedSetterProperty', 'ApiInputs',
           'DiscoveryIntegrationInput', 'DataFeedFileProcessOutput', 'requests_retry_session', 'is_valid_uuid', 'convert_bytes']

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


class ValidatedSetterProperty:
    def __init__(self, func, name=None, doc=None):
        self.func = func
        self.__name__ = name if name is not None else func.__name__
        self.__doc__ = doc if doc is not None else func.__doc__

    def __set__(self, obj, value):
        ret = self.func(obj, value)
        obj.__dict__[self.__name__] = value


def _column_name_cap(columns) -> list:
    renamed_columns = [name[0].upper() + name[1:] for name in columns.to_list()]
    return renamed_columns


def _with_func_attrs(**attrs):
    def with_attrs(f):
        for k, v in attrs.items():
            setattr(f, k, v)
        return f

    return with_attrs


def _is_valid_regex(regex):
    """Check if regex is valid.

    Parameters
    ----------
    regex : str
        A `regex` expression to be validated.

    Returns
    -------
    bool
        Valid if True, invalid if False.

    """
    try:
        re.compile(regex)
        return True
    except re.error:
        return False


ApiHeaders = namedtuple('ApiHeaders', ['default', 'integration'])

ApiInputs = namedtuple('ApiInputs',
                       ['email_address', 'user_id', 'api_project_id', 'data_feed_id', 'data_feed_file_status_id',
                        'bearer_token', 'api_base_url', 'api_projects_endpoint', 'subscription_key', 'api_headers'])

IngestionMode = namedtuple('IngestionMode', ['Queue', 'Stream'])

DiscoveryIntegrationInput = namedtuple(
    'DiscoveryIntegrationInput',
    ['api_project_id', 'installation_id', 'network_device_id', 'integration_device_id', 'user_id', 'batch_no', 'job_id'])
# DiscoveryIntegrationInput.__doc__ = """DiscoveryIntegrationInput(api_project_id, installation_id, network_device_id, integration_device_id, user_id, batch_no)
#
# Defines the required inputs to be provided to the run_discovery() method's integration_input parameter.
#
# Parameters
# ----------
# api_project_id : the unique identifier for the portfolio (ApiProjectID)
# installation_id : the unique identifier for the given site (InstallationID)
# network_device_id : the DeviceSpecs.NetworkDeviceID for the given integration_device_id
# integration_device_id : the unique identifier for the integration device the discovery is triggered against
# user_id : the unique identifier of the user who triggered the discovery
# batch_no : the unique identifier for the instance of the triggered discovery
# """
# DiscoveryIntegrationInput.api_project_id.__doc__ = "uuid.UUID : the unique identifier for the portfolio (ApiProjectID)"
# DiscoveryIntegrationInput.installation_id.__doc__ = "uuid.UUID : the unique identifier for the given site (InstallationID)"
# DiscoveryIntegrationInput.network_device_id.__doc__ = "uuid.UUID : the DeviceSpecs.NetworkDeviceID for the given integration_device_id"
# DiscoveryIntegrationInput.integration_device_id.__doc__ = "uuid.UUID : the unique identifier for the integration device the discovery is triggered against"
# DiscoveryIntegrationInput.user_id.__doc__ = "uuid.UUID : the unique identifier of the user who triggered the discovery"
# DiscoveryIntegrationInput.batch_no.__doc__ = "uuid.UUID : the unique identifier for the instance of the triggered discovery"

DataFeedFileProcessOutput = namedtuple('DataFeedFileProcessOutput',
                                       [
                                           'data_feed_id', 'data_feed_file_status_id',
                                           'client_tracking_id',
                                           'source_type', 'file_name', 'file_received',
                                           'file_process_status',
                                           'file_process_status_change', 'process_started',
                                           'process_completed',
                                           'minutes_since_received', 'minutes_since_processed',
                                           'file_size',
                                           'log_file_path', 'output', 'error'
                                       ])


def generate_password():
    """Generate a ten-character alphanumeric password with at least one lowercase character, at least one uppercase
    character, and at least three digits.
    """
    alphabet = string.ascii_letters + string.digits
    while True:
        password = ''.join(secrets.choice(alphabet) for i in range(10))
        if (any(c.islower() for c in password)
                and any(c.isupper() for c in password)
                and sum(c.isdigit() for c in password) >= 3):
            break
    return password


def convert_to_sqm(df: pandas.DataFrame, sqft_col_name: str):
    """Convert to square metres

    Convert floor area from square feet to square metres.

    Parameters
    ----------
    df : pandas.DataFrame
        The site asset register dataframe.
    sqft_col_name : str
        The name of the column containing the Floor Area values in sq. ft. to be converted.

    Returns
    -------
    df : pandas.DataFrame
        The input dataframe after converting the `sqft_col_name` column to square metres.


    """
    df[sqft_col_name] = df[sqft_col_name] * 0.09290304
    return df


def convert_to_pascal_case(text: Union[str, List[str]]):
    """Convert to PascalCase

    Capitalises the first letter of each word and strips any non-alphanumeric (``[a-zA-Z0-9]``) characters from between
    words as well as any leading or trailing the string.

    Parameters
    ----------
    text : Union[str, List[str]]
        A string or list of strings

    Returns
    -------
    Union[str, List[str]]
        If passed a string, the function returns the converted string. If passed a list of strings, the function
        returns a list of converted strings.

    """
    if type(text) == str:
        text = re.sub(r"^[^a-zA-Z0-9]", '', str(text))
        text = str.upper(text[0]) + text[1:]
        if not text:
            return text
        text = re.sub(r"[^a-zA-Z0-9]([a-zA-Z0-9])", lambda matched: str.upper(matched.group(1)), text)
        return re.sub(r"[^a-zA-Z0-9]", '', text)
    elif type(text) == list and {isinstance(text[j], str) for j in range(len(text))}.issubset([True]):
        for i in range(len(text)):
            text_item = text[i]

            text_item = re.sub(r"^[^a-zA-Z0-9]", '', str(text_item))
            text_item = str.upper(text_item[0]) + text_item[1:]
            if not text_item:
                text[i] = text_item
            text_item = re.sub(r"[^a-zA-Z0-9]([a-zA-Z0-9])", lambda matched: str.upper(matched.group(1)), text_item)

            text[i] = re.sub(r"[^a-zA-Z0-9]", '', text_item)
        return text
    else:
        logger.exception(f"The value passed to the text parameter was not a string or a list of strings. "
                         f"Unable to convert to pascal case. ")
        return False


def requests_retry_session(retries=5, status_forcelist=[429, 500, 502, 503, 504], backoff_factor=0.3, session=None, method_whitelist=["GET"]):
    """
    Provide retry policy for requests fired.

    Parameters
    ----------
    retries : number
        Number of retries the retry would do before declaring it as an exception.
        Defaults to 5.
    status_forcelist : List[number]
        A set of integer HTTP status codes that we should force a retry on.
        Defaults to 429, 500, 502, 503, 504.
    backoff_factor : number
        A backoff factor to apply between attempts after the second try (most errors are resolved immediately by a second try without a delay).
        {backoff factor} * (2 ** ({number of total retries} - 1)) seconds
        Defaults to 0.3
    method_whitelist : List[string]
        Set of uppercased HTTP method verbs that we should retry on
        Defaults to: GET

    Returns
    -------
        Session containing the retry configuration.
        Can be changed as requests_retry_session().get() or .post() depending on requirements
    """
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        status_forcelist=status_forcelist,
        backoff_factor=backoff_factor,
        method_whitelist=method_whitelist
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter=adapter)
    session.mount('https://', adapter=adapter)
    return session

def is_valid_uuid(uuid_to_test: Union[uuid.UUID, str], version=4) -> bool:
    """
    Check if uuid_to_test is a valid UUID.
    
     Parameters
    ----------
    uuid_to_test : str
    version : {1, 2, 3, 4}
    
     Returns
    -------
    `True` if uuid_to_test is a valid UUID, otherwise `False`.
    """

    if uuid_to_test is None:
        return False

    if not isinstance(uuid_to_test, uuid.UUID) and not isinstance(uuid_to_test, str):
        print('uuid_to_test is not a valid UUID or str type.')
        return False

    try:
        if isinstance(uuid_to_test, uuid.UUID):

            if uuid_to_test == uuid.UUID(int=0):
                return False
            
            return uuid_to_test.version == version
         
        uuid_obj = uuid.UUID(str(uuid_to_test), version=version)
    except Exception:
        return False

    return str(uuid_obj) == uuid_to_test

def convert_bytes(bytes_number):
    tags = [ "Byte", "Kilobyte", "Megabyte", "Gigabyte", "Terabyte" ]
 
    i = 0
    double_bytes = bytes_number
 
    while (i < len(tags) and  bytes_number >= 1024):
            double_bytes = bytes_number / 1024.0
            i = i + 1
            bytes_number = bytes_number / 1024
 
    return str(round(double_bytes, 2)) + " " + tags[i]

_site_schema = pandera.DataFrameSchema(
    {
        'InstallationName': pandera.Column(pandera.String, checks=pandera.Check.str_length(1, 255)),
        'InstallationCode': pandera.Column(pandera.STRING, coerce=True, checks=pandera.Check.str_length(1, 100)),
        'Address': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 250)]),
        'Country': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 50)]),
        'Suburb': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 50)]),
        'State': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 100)], required=False),
        'StateName': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 50, n_failure_cases=None)]),
        'FloorAreaM2': pandera.Column(pandera.Float, checks=[pandera.Check.greater_than_or_equal_to(0)]),
        'ZipPostCode': pandera.Column(pandera.STRING, checks=[pandera.Check.str_length(1, 20)]),
        'Latitude': pandera.Column(pandera.Float, required=False),
        'Longitude': pandera.Column(pandera.Float, required=False),
        'Timezone': pandera.Column(pandera.String, required=False),
        'InstallationId': pandera.Column(pandera.String, required=False),
    }
)

_work_order_schema = pandera.DataFrameSchema(
    columns={
        'WorkOrderId': pandera.Column(pandera.String, required=True),
        'InstallationId': pandera.Column(pandera.String, checks=[
            pandera.Check(lambda x: uuid.UUID(x).__class__ == uuid.UUID, element_wise=True)], required=True),
        'WorkOrderSiteIdentifier': pandera.Column(pandera.STRING, coerce=True, required=True),
        'WorkOrderCategory': pandera.Column(pandera.String, checks=[
            pandera.Check.isin(list(WORK_ORDER_CATEGORY.__args__), ignore_na=False)], required=True),
        'Type': pandera.Column(pandera.String, required=True),
        'Description': pandera.Column(pandera.String, required=True),
        'Priority': pandera.Column(pandera.String, checks=[
            pandera.Check.isin(list(WORK_ORDER_PRIORITY.__args__), ignore_na=False)], required=True),
        'Status': pandera.Column(pandera.String, checks=[
            pandera.Check.isin(list(WORK_ORDER_STATUS.__args__), ignore_na=False)], required=True),
        'CreatedDate': pandera.Column(pandera.DateTime, required=True, nullable=False),
        'LastModifiedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'WorkStartedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'WorkCompletedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'ClosedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'RawPriority': pandera.Column(pandera.String, required=True, nullable=False),
        'RawWorkOrderCategory': pandera.Column(pandera.String, required=True, nullable=False),
        'SubType': pandera.Column(pandera.String, required=False, nullable=True),
        'Vendor': pandera.Column(pandera.String, required=False, nullable=True),
        'VendorId': pandera.Column(pandera.String, required=False, nullable=True),
        'EquipmentClass': pandera.Column(pandera.String, required=False, nullable=True),
        'RawEquipmentClass': pandera.Column(pandera.String, required=False, nullable=True),
        'EquipmentLabel': pandera.Column(pandera.String, required=False, nullable=True),
        'RawEquipmentId': pandera.Column(pandera.String, required=False, nullable=True),
        'TenantId': pandera.Column(pandera.String, required=False, nullable=True),
        'TenantName': pandera.Column(pandera.String, required=False, nullable=True),
        'NotToExceedCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'TotalCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'BillableCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'NonBillableCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'Location': pandera.Column(pandera.String, required=False, nullable=True),
        'RawLocation': pandera.Column(pandera.String, required=False, nullable=True),
        'ScheduledStartDate': pandera.Column(pandera.DateTime, required=False, nullable=True),
        'ScheduledCompletionDate': pandera.Column(pandera.DateTime, required=False, nullable=True),
    },
    coerce=True,
    strict=False,
    name=None
)
